# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


from __future__ import annotations
from scipy.sparse import csr_matrix # type: ignore
from scipy.sparse.csgraph import reverse_cuthill_mckee # type: ignore
from scipy.sparse.linalg import bicgstab, gmres, gcrotmk, eigs, splu # type: ignore
from scipy.linalg import eig # type: ignore
from scipy import sparse # type: ignore
from dataclasses import dataclass, field
import numpy as np
from loguru import logger
import platform
import time
from typing import Literal, Callable
from enum import Enum

_PARDISO_AVAILABLE = False
_UMFPACK_AVAILABLE = False
_CUDSS_AVAILABLE = False

""" Check if the PC runs on a non-ARM architechture
If so, attempt to import PyPardiso (if its installed)
"""


############################################################
#                          PARDISO                         #
############################################################

if 'arm' not in platform.processor():
    try:
        from .solve_interfaces.pardiso_interface import PardisoInterface
        _PARDISO_AVAILABLE = True
    except ModuleNotFoundError:
        logger.info('Pardiso not found, defaulting to SuperLU')


############################################################
#                          UMFPACK                         #
############################################################


try:
    import scikits.umfpack as um # type: ignore
    _UMFPACK_AVAILABLE = True
except ModuleNotFoundError:
    logger.debug('UMFPACK not found, defaulting to SuperLU')


############################################################
#                           CUDSS                          #
############################################################

try:
    from .solve_interfaces.cudss_interface import CuDSSInterface
    _CUDSS_AVAILABLE = True
except ModuleNotFoundError:
    pass

############################################################
#                       SOLVE REPORT                       #
############################################################

@dataclass
class SolveReport:
    simtime: float = -1.0
    jobid: int = -1
    ndof: int = -1
    nnz: int = -1
    ndof_solve: int = -1
    nnz_solve: int = -1
    exit_code: int = 0
    solver: str = 'None'
    sorter: str = 'None'
    precon: str = 'None'
    aux: dict[str, str] = field(default_factory=dict)

    def add(self, **kwargs: str):
        for key, value in kwargs.items():
            self.aux[key] = str(value)
    
    def pretty_print(self, print_cal: Callable | None = None):
        if print_cal is None:
            print_cal = print
        # Set column widths
        col1_width = 22  # Wider key column
        col2_width = 40  # Value column
        total_width = col1_width + col2_width + 5  # +5 for borders/padding

        def row(key, val):
            val_str = f"{val:.4f}" if isinstance(val, float) else str(val)
            print_cal(f"| {key:<{col1_width}} | {val_str:<{col2_width}} |") # ty: ignore

        border = "+" + "-" * (col1_width + 2) + "+" + "-" * (col2_width + 2) + "+"

        print_cal(border)
        print_cal(f"| {'FEM Solve Report':^{total_width - 2}} |")
        print_cal(border)
        row("Solver", self.solver)
        row("Sorter", self.sorter)
        row("Preconditioner", self.precon)
        row("Job ID", self.jobid)
        row("Sim Time (s)", self.simtime)
        row("DOFs (Total)", self.ndof)
        row("NNZ (Total)", self.nnz)
        row("DOFs (Solve)", self.ndof_solve)
        row("NNZ (Solve)", self.nnz_solve)
        row("Exit Code", self.exit_code)
        print_cal(border)

        if self.aux:
            print_cal(f"| {'Additional Info':^{total_width - 2}} |")
            print_cal(border)
            for k, v in self.aux.items():
                row(str(k), v)
            print_cal(border)

############################################################
#                 EIGENMODE FILTER ROUTINE                #
############################################################

def filter_real_modes(eigvals: np.ndarray, eigvecs: np.ndarray, 
                      k0: float, ermax: complex, urmax: complex, sign: float) -> tuple[np.ndarray, np.ndarray]:
    """
    Given arrays of eigenvalues `eigvals` and eigenvectors `eigvecs` (cols of shape (N,)),
    and a free‐space wavenumber k0, return only those eigenpairs whose eigenvalue can
    correspond to a real propagation constant β (i.e. 0 ≤ β² ≤ k0²·ermax·urmax).

    Assumes that `ermax` and `urmax` are defined in the surrounding scope.

    Parameters
    ----------
    eigvals : 1D array_like of float
        The generalized eigenvalues (β² candidates).
    eigvecs : 2D array_like, shape (N, M)
        The corresponding eigenvectors, one column per eigenvalue.
    k0 : float
        Free‐space wavenumber.

    Returns
    -------
    filtered_vals : 1D ndarray
        Subset of `eigvals` satisfying 0 ≤ eigval ≤ k0²·ermax·urmax (within numerical tol).
    filtered_vecs : 2D ndarray
        Columns of `eigvecs` corresponding to `filtered_vals`.
    """
    minimum = 1
    extremum = (k0**2) * ermax * urmax * 2
    
    mask = (sign*eigvals <= extremum) & (sign*eigvals >= minimum)
    filtered_vals = eigvals[mask]
    filtered_vecs = eigvecs[:, mask]
    k0vals = np.sqrt(sign*filtered_vals)
    order = np.argsort(np.abs(k0vals))# ascending distance
    filtered_vals = filtered_vals[order]             # reorder eigenvalues
    filtered_vecs = filtered_vecs[:, order] 
    return filtered_vals, filtered_vecs


############################################################
#               EIGENMODE ORTHOGONALITY CHECK              #
############################################################

def filter_unique_eigenpairs(eigen_values: list[complex], eigen_vectors: list[np.ndarray], tol=-3) -> tuple[list[complex], list[np.ndarray]]:
    """
    Filters eigenvectors by orthogonality using dot-product tolerance.
    
    Parameters:
        eigen_values (np.ndarray): Array of eigenvalues, shape (n,)
        eigen_vectors (np.ndarray): Array of eigenvectors, shape (n, n)
        tol (float): Dot product tolerance for considering vectors orthogonal (default: 1e-5)

    Returns:
        unique_values (np.ndarray): Filtered eigenvalues
        unique_vectors (np.ndarray): Corresponding orthogonal eigenvectors
    """
    selected: list = []
    indices: list = []
    for i in range(len(eigen_vectors)):
        
        vec = eigen_vectors[i]
        vec = vec / np.linalg.norm(vec)  # Normalize

        # Check orthogonality against selected vectors
        if all(10*np.log10(abs(np.dot(vec, sel))) < tol for sel in selected):
            selected.append(vec)
            indices.append(i)

    unique_values = [eigen_values[i] for i in indices]
    unique_vectors = [eigen_vectors[i] for i in indices]

    return unique_values, unique_vectors


############################################################
#         COMPLEX MATRIX TO REAL MATRIX CONVERSION        #
############################################################

def complex_to_real_block(A, b):
    """Return (Â,  b̂) real-augmented representation of A x = b."""
    A_r = sparse.csr_matrix(A.real)
    A_i = sparse.csr_matrix(A.imag)
    #  [ ReA  -ImA ]
    #  [ ImA   ReA ]
    upper = sparse.hstack([A_r, -A_i])
    lower = sparse.hstack([A_i,  A_r])
    A_hat = sparse.vstack([upper, lower]).tocsr()

    b_hat = np.hstack([b.real, b.imag])
    return A_hat, b_hat

def real_to_complex_block(x):
    """Return x = (x_r, x_i) as complex vector."""
    n = x.shape[0] // 2
    x_r = x[:n]
    x_i = x[n:]
    return x_r + 1j * x_i


############################################################
#                  BASE CLASS DEFINITIONS                 #
############################################################

class SimulationError(Exception):
    pass

class Sorter:
    """ A Generic class that executes a sort on the indices.
    It must implement a sort and unsort method.
    """
    def __init__(self):
        self.perm = None
        self.inv_perm = None

    def reset(self):
        """ Reset the permuation vectors."""
        self.perm = None
        self.inv_perm = None

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def sort(self, A: csr_matrix, b: np.ndarray, reuse_sorting: bool = False) -> tuple[csr_matrix, np.ndarray]:
        return A,b
    
    def unsort(self, x: np.ndarray) -> np.ndarray:
        return x

class Preconditioner:
    """A Generic class defining a preconditioner as attribute .M based on the
    matrix A and b. This must be generated in the .init(A,b) method.
    """
    def __init__(self):
        self.M: np.ndarray = None

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def init(self, A: csr_matrix, b: np.ndarray) -> None:
        raise NotImplementedError('')

class Solver:
    """ A generic class representing a solver for the problem Ax=b
    
    A solver class has two class attributes.
     - real_only: defines if the solver can only deal with real numbers. In this case
    the solve routine will automatically provide A and b in real number format.
     - req_sorter: defines if this solver requires the use of a sorter algorithm. By setting
     it to False, the SolveRoutine will not use the default sorting algorithm.
    """
    real_only: bool = False
    req_sorter: bool = False

    def __init__(self):
        self.own_preconditioner: bool = False

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def duplicate(self) -> Solver:
        return self.__class__()

    def set_options(self, pivoting_threshold: float | None = None) -> None:
        """Write generic simulation options to the solver object. 
        Options may be ignored depending on the type of solver used."""
        pass
        
    def solve(self, A: csr_matrix, b: np.ndarray, precon: Preconditioner, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        raise NotImplementedError("This classes Ax=B solver method is not implemented.")
    
    def reset(self) -> None:
        """Reset state variables like numeric and symbollic factorizations."""
        pass

class EigSolver:
    """ A generic class representing a solver for the eigenvalue problem Ax=λBx
    
    A solver class has two class attributes.
     - real_only: defines if the solver can only deal with real numbers. In this case
    the solve routine will automatically provide A and b in real number format.
     - req_sorter: defines if this solver requires the use of a sorter algorithm. By setting
     it to False, the SolveRoutine will not use the default sorting algorithm.
    """
    real_only: bool = False
    req_sorter: bool = False

    def __init__(self):
        self.own_preconditioner: bool = False

    def __str__(self) -> str:
        return f'{self.__class__.__name__}'
    
    def duplicate(self) -> Solver:
        return self.__class__()

    def eig(self, A: csr_matrix | csr_matrix, B: csr_matrix | csr_matrix, nmodes: int = 6, target_k0: float = 0.0, which: str = 'LM', sign: float = 1.):
        raise NotImplementedError("This classes eigenmdoe solver method is not implemented.")
    
    def reset(self) -> None:
        """Reset state variables like numeric and symbollic factorizations."""
        pass



############################################################
#                          SORTERS                         #
############################################################


class ReverseCuthillMckee(Sorter):
    """ Implements the Reverse Cuthill-Mckee sorting."""
    def __init__(self):
        super().__init__()
        

    def sort(self, A, b, reuse_sorting: bool = False):
        if not reuse_sorting:
            logger.debug('Generating Reverse Cuthill-Mckee sorting.')
            self.perm = reverse_cuthill_mckee(A)
            self.inv_perm = np.argsort(self.perm)
        logger.debug('Applying Reverse Cuthill-Mckee sorting.')
        Asorted = A[self.perm, :][:, self.perm]
        bsorted = b[self.perm]
        return Asorted, bsorted
    
    def unsort(self, x: np.ndarray):
        logger.debug('Reversing Reverse Cuthill-Mckee sorting.')
        return  x[self.inv_perm]
    

############################################################
#                      PRECONDITIONERS                     #
############################################################


class ILUPrecon(Preconditioner):
    """ Implements the incomplete LU preconditioner on matrix A. """
    def __init__(self):
        super().__init__()
        self.M = None
        self.fill_factor = 10
        self.options: dict[str, str] = dict(SymmetricMode=True)

    def init(self, A, b):
        logger.info("Generating ILU Preconditioner")
        self.ilu = sparse.linalg.spilu(A, drop_tol=1e-2, fill_factor=self.fill_factor, permc_spec='MMD_AT_PLUS_A', diag_pivot_thresh=0.001, options=self.options) # ty: ignore
        self.M = sparse.linalg.LinearOperator(A.shape, self.ilu.solve) # ty: ignore


############################################################
#                     ITERATIVE SOLVERS                    #
############################################################


class SolverBicgstab(Solver):
    """ Implements the Bi-Conjugate Gradient Stabilized method"""
    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, xk):
        convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {convergence:.4f}')

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling BiCGStab.')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = bicgstab(A, b, M=precon.M, atol=self.atol, callback=self.callback)
        else:
            x, info = bicgstab(A, b, atol=self.atol, callback=self.callback)
        return x, SolveReport(solver=str(self), exit_code=info)

class SolverGCROTMK(Solver):
    """ Implements the GCRO-T(m,k) Iterative solver. """
    def __init__(self):
        super().__init__()
        self.atol = 1e-5
        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, xk):
        convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {convergence:.4f}')

    def solve(self, A: csr_matrix, b: np.ndarray, precon: Preconditioner, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling GCRO-T(m,k) algorithm')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = gcrotmk(A, b, M=precon.M, atol=self.atol, callback=self.callback)
        else:
            x, info = gcrotmk(A, b, atol=self.atol, callback=self.callback)
        return x, SolveReport(solver=str(self), exit_code=info)

class SolverGMRES(Solver):
    """ Implements the GMRES solver. """
    real_only = False
    req_sorter = True

    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def callback(self, norm):
        #convergence = np.linalg.norm((self.A @ xk - self.b))
        logger.info(f'Iteration {norm:.4f}')

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling GMRES Function.')
        self.A = A
        self.b = b
        if precon.M is not None:
            x, info = gmres(A, b, M=precon.M, atol=self.atol, callback=self.callback, callback_type='pr_norm')
        else:
            x, info = gmres(A, b, atol=self.atol, callback=self.callback, restart=500, callback_type='pr_norm')
        return x, SolveReport(solver=str(self), exit_code=info)


############################################################
#                      DIRECT SOLVERS                     #
############################################################


class SolverSuperLU(Solver):
    """ Implements Scipi's direct SuperLU solver."""
    req_sorter: bool = False
    real_only: bool = False
    
    def __init__(self):
        super().__init__()
        self.atol = 1e-5

        self.A: np.ndarray = None
        self.b: np.ndarray = None
        self.options: dict[str, str] = dict(SymmetricMode=True, Equil=False, IterRefine='SINGLE')
        self._pivoting_threshold: float = 0.001
        self.lu = None
    
    def duplicate(self) -> Solver:
        new_solver = self.__class__()
        new_solver._pivoting_threshold = self._pivoting_threshold
        return new_solver

    def set_options(self,
                    pivoting_threshold: float | None = None) -> None:
        if pivoting_threshold is not None:
            self._pivoting_threshold = pivoting_threshold
    
    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling SuperLU Solver.')
        self.single = True
        if not reuse_factorization:
            self.lu = splu(A, permc_spec='MMD_AT_PLUS_A', relax=0, diag_pivot_thresh=self._pivoting_threshold, options=self.options)
        x = self.lu.solve(b)
        aux = {
            "pivoting threshold": str(self._pivoting_threshold)
        }
        return x, SolveReport(solver=str(self), exit_code=0, aux=aux)

class SolverUMFPACK(Solver):
    """ Implements the UMFPACK Sparse SP solver."""
    req_sorter = False
    real_only = False

    def __init__(self):
        super().__init__()
        self.A: np.ndarray = None
        self.b: np.ndarray = None
        self.umfpack: um.UmfpackContext = um.UmfpackContext('zl')
        self.umfpack.control[um.UMFPACK_PRL] = 0 # ty: ignore
        self.umfpack.control[um.UMFPACK_IRSTEP] = 2 # ty: ignore
        self.umfpack.control[um.UMFPACK_STRATEGY] = um.UMFPACK_STRATEGY_SYMMETRIC # ty: ignore
        self.umfpack.control[um.UMFPACK_ORDERING] = 3 # ty: ignore
        self.umfpack.control[um.UMFPACK_PIVOT_TOLERANCE] = 0.001 # ty: ignore
        self.umfpack.control[um.UMFPACK_SYM_PIVOT_TOLERANCE] = 0.001 # ty: ignore
        self.umfpack.control[um.UMFPACK_BLOCK_SIZE] = 64 # ty: ignore
        self.umfpack.control[um.UMFPACK_FIXQ] = -1 # ty: ignore
        
        # SETTINGS
        self._pivoting_threshold: float = 0.001

        self.fact_symb: bool = False

    def reset(self) -> None:
        self.fact_symb = False
    
    def set_options(self,
                    pivoting_threshold: float | None = None) -> None:
            if pivoting_threshold is not None:
                self.umfpack.control[um.UMFPACK_PIVOT_TOLERANCE] = pivoting_threshold # ty: ignore
                self.umfpack.control[um.UMFPACK_SYM_PIVOT_TOLERANCE] = pivoting_threshold # ty: ignore
                self._pivoting_threshold = pivoting_threshold

    def duplicate(self) -> Solver:
        new_solver = self.__class__()
        new_solver.set_options(pivoting_threshold = self._pivoting_threshold)
        return new_solver

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling UMFPACK Solver.')
        A.indptr  = A.indptr.astype(np.int64)
        A.indices = A.indices.astype(np.int64)
        if self.fact_symb is False:
            logger.debug(f'[ID={id}] Executing symbollic factorization.')
            self.umfpack.symbolic(A)
            self.fact_symb = True
        if not reuse_factorization:
            #logger.debug('Executing numeric factorization.')
            self.umfpack.numeric(A)
            self.A = A
        x = self.umfpack.solve(um.UMFPACK_A, self.A, b, autoTranspose = False ) # ty: ignore
        aux = {
            "Pivoting Threshold": str(self._pivoting_threshold),
        }
        return x, SolveReport(solver=str(self), exit_code=0, aux=aux)

class SolverPardiso(Solver):
    """ Implements the PARDISO solver through PyPardiso. """
    real_only: bool = False
    req_sorter: bool = False

    def __init__(self):
        super().__init__()
        self.solver: PardisoInterface = PardisoInterface()
        self.fact_symb: bool = False
        self.A: np.ndarray = None
        self.b: np.ndarray = None

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1) -> tuple[np.ndarray, SolveReport]:
        logger.info(f'[ID={id}] Calling Pardiso Solver')
        if self.fact_symb is False:
            logger.debug(f'[ID={id}] Executing symbollic factorization.')
            self.solver.symbolic(A)
            self.fact_symb = True
        if not reuse_factorization:
            self.solver.numeric(A)
            self.A = A
        x, error = self.solver.solve(A, b)
        if error != 0:
            logger.error(f'[ID={id}] Terminated with error code {error}')
            logger.error(self.solver.get_error(error))
            raise SimulationError(f'[ID={id}] PARDISO Terminated with error code {error}')
        aux = {}
        return x, SolveReport(solver=str(self), exit_code=error, aux=aux)
    

class CuDSSSolver(Solver):
    real_only = False
    def __init__(self):
        self._cudss = CuDSSInterface()
        self.fact_symb: bool = False
        self.fact_numb: bool = False
        self._cudss._PRES = 2

    def reset(self) -> None:
        self.fact_symb = False
        self.fact_numb = False

    def solve(self, A, b, precon, reuse_factorization: bool = False, id: int = -1):
        logger.info(f'[{id}] Calling cuDSS Solver')
        
        if self.fact_symb is False:
            logger.debug('Executing symbollic factorization')
            x = self._cudss.from_symbolic(A,b)
            self.fact_symb = True
        else:
            if reuse_factorization:
                x = self._cudss.from_solve(b)
            else:
                x = self._cudss.from_numeric(A,b)
        
        return x, SolveReport(solver=str(self), exit_code=0, aux={})


############################################################
#                 DIRECT EIGENMODE SOLVERS                #
############################################################

class SolverLAPACK(EigSolver):

    def __init__(self):
        super().__init__()
    
    def eig(self, 
            A: csr_matrix | csr_matrix, 
            B: csr_matrix | csr_matrix,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.0) -> tuple[np.ndarray, np.ndarray]:
        """
        Dense solver for  A x = λ B x   with A = Aᴴ, B = Bᴴ (B may be indefinite).

        Parameters
        ----------
        A, B : (n, n) array_like, complex127/complex64/float64
        k    : int or None
            How many eigenpairs to return.
            * None  → return all n
            * k>0   → return k pairs with |λ| smallest

        Returns
        -------
        lam  : (m,) real ndarray      eigenvalues  (m = n or k)
        vecs : (n, m) complex ndarray eigenvectors, B-orthonormal  (xiᴴ B xj = δij)
        """
        logger.debug('Calling LAPACK eig solver')
        lam, vecs = eig(A.toarray(), B.toarray(), overwrite_a=True, overwrite_b=True, check_finite=False)
        lam, vecs = filter_real_modes(lam, vecs, target_k0, 2, 2, sign=sign)
        return lam, vecs
    

############################################################
#                  ITERATIVE EIGEN SOLVERS                 #
############################################################


class SolverARPACK(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver."""
    def __init__(self):
        super().__init__()

    def eig(self, 
            A: csr_matrix | csr_matrix, 
            B: csr_matrix | csr_matrix,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.0) -> tuple[np.ndarray, np.ndarray]:
        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        sigma = sign*(target_k0**2)
        eigen_values, eigen_modes = eigs(A, k=nmodes, M=B, sigma=sigma, which=which)
        return eigen_values, eigen_modes

class SmartARPACK_BMA(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver with automatic search.
    
    The Solver searches in a geometric range around the target wave constant.
    """
    def __init__(self):
        super().__init__()
        self.symmetric_steps: int = 41
        self.search_range: float = 2.0
        self.energy_limit: float = 1e-4

    def eig(self, 
            A: csr_matrix | csr_matrix, 
            B: csr_matrix | csr_matrix,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.) -> tuple[np.ndarray, np.ndarray]:

        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        qs = np.geomspace(1, self.search_range, self.symmetric_steps)
        tot_eigen_values = []
        tot_eigen_modes = []
        energies = []
        for i, q in enumerate(qs):
            # Above target k0
            sigma = sign*((q*target_k0)**2)
            eigen_values, eigen_modes = eigs(A, k=1, M=B, sigma=sigma, which=which)
            energy = np.mean(np.abs(eigen_modes.flatten())**2)
            if energy > self.energy_limit:
                tot_eigen_values.append(eigen_values[0])
                tot_eigen_modes.append(eigen_modes.flatten())
                energies.append(energy)
            if i!=0:
                # Below target k0
                sigma = sign*((target_k0/q)**2)
                eigen_values, eigen_modes = eigs(A, k=1, M=B, sigma=sigma, which=which)
                energy = np.mean(np.abs(eigen_modes.flatten())**2)
                if energy > self.energy_limit:
                    tot_eigen_values.append(eigen_values[0])
                    tot_eigen_modes.append(eigen_modes.flatten())
                    energies.append(energy)
        
        #Sort solutions on mode energy
        val, mode, energy = zip(*sorted(zip(tot_eigen_values,tot_eigen_modes,energies), key=lambda x: x[2], reverse=True))
        eigen_values = np.array(val[:nmodes])
        eigen_modes = np.array(mode[:nmodes]).T

        return eigen_values, eigen_modes
    
class SmartARPACK(EigSolver):
    """ Implements the Scipy ARPACK iterative eigenmode solver with automatic search.
    
    The Solver searches in a geometric range around the target wave constant.
    """
    def __init__(self):
        super().__init__()
        self.symmetric_steps: int = 3
        self.search_range: float = 2.0
        self.energy_limit: float = 1e-4

    def eig(self, 
            A: csr_matrix | csr_matrix, 
            B: csr_matrix | csr_matrix,
            nmodes: int = 6,
            target_k0: float = 0,
            which: str = 'LM',
            sign: float = 1.) -> tuple[np.ndarray, np.ndarray]:
        logger.info(f'Searching around 	β = {target_k0:.2f} rad/m')
        qs = np.geomspace(1, self.search_range, self.symmetric_steps)
        tot_eigen_values = []
        tot_eigen_modes = []
        for i, q in enumerate(qs):
            # Above target k0
            sigma = sign*((q*target_k0)**2)
            eigen_values, eigen_modes = eigs(A, k=6, M=B, sigma=sigma, which=which)
            for j in range(eigen_values.shape[0]):
                if eigen_values[j]<(sigma/self.search_range):
                    continue
                tot_eigen_values.append(eigen_values[j])
                tot_eigen_modes.append(eigen_modes[:,j])
            if i!=0:
                # Below target k0
                sigma = sign*((target_k0/q)**2)
                eigen_values, eigen_modes = eigs(A, k=6, M=B, sigma=sigma, which=which)
                for j in range(eigen_values.shape[0]):
                    if eigen_values[j]<(sigma/self.search_range):
                        continue
                    tot_eigen_values.append(eigen_values[j])
                    tot_eigen_modes.append(eigen_modes[:,j])
            tot_eigen_values, tot_eigen_modes = filter_unique_eigenpairs(tot_eigen_values, tot_eigen_modes)
            if len(tot_eigen_values)>nmodes:
                break
        #Sort solutions on mode energy
        val, mode = filter_unique_eigenpairs(tot_eigen_values, tot_eigen_modes)
        val, mode = zip(*sorted(zip(val,mode), key=lambda x: x[0], reverse=False)) # type: ignore
        eigen_values = np.array(val[:nmodes])
        eigen_modes = np.array(mode[:nmodes]).T

        return eigen_values, eigen_modes



############################################################
#                        SOLVER ENUM                       #
############################################################


class EMSolver(Enum):
    SUPERLU = 1
    UMFPACK = 2
    PARDISO = 3
    LAPACK = 4
    ARPACK = 5
    SMART_ARPACK = 6
    SMART_ARPACK_BMA = 7
    CUDSS = 8

    def get_solver(self) -> Solver | EigSolver | None:
        if self==EMSolver.SUPERLU:
            return SolverSuperLU()
        elif self==EMSolver.UMFPACK:
            if _UMFPACK_AVAILABLE is False:
                return None
            else:
                return SolverUMFPACK()
        elif self==EMSolver.PARDISO:
            if _PARDISO_AVAILABLE is False:
                return None
            else:
                return SolverPardiso()
        elif self==EMSolver.LAPACK:
            return SolverLAPACK()
        elif self==EMSolver.ARPACK:
            return SolverARPACK()
        elif self==EMSolver.SMART_ARPACK:
            return SmartARPACK()
        elif self==EMSolver.SMART_ARPACK_BMA:
            return SmartARPACK_BMA()
        elif self==EMSolver.CUDSS:
            if _CUDSS_AVAILABLE is False:
                return None
            else:
                return CuDSSSolver()
        raise ValueError(f'An unsupported Enum case has been reached: {self}')

############################################################
#                       SOLVE ROUTINE                      #
############################################################


class SolveRoutine:
    """ A generic class describing a solve routine.
    A solve routine contains all the relevant sorter preconditioner and solver objects
    and goes through a sequence of steps to solve a linear system or find eigenmodes.

    """
    def __init__(self):
        
        self.sorter: Sorter = ReverseCuthillMckee()
        self.precon: Preconditioner = ILUPrecon()
        self.solvers: dict[EMSolver, Solver | EigSolver] = {slv: slv.get_solver() for slv in EMSolver}
        self.solvers = {key: solver for key, solver in self.solvers.items() if solver is not None}

        self.parallel: Literal['SI','MT','MP'] = 'SI'
        self.smart_search: bool = False
        self.forced_solver: list[Solver | EigSolver] = []
        self.disabled_solver: list[type[Solver]|type[EigSolver]] = []

        self.use_sorter: bool = False
        self.use_preconditioner: bool = False
        self.use_direct: bool = True

    
    def __str__(self) -> str:
        return 'SolveRoutine()'
    
    def _legal_solver(self, solver: Solver | EigSolver) -> bool:
        """Checks if a solver is a legal option.

        Args:
            solver (Solver): The solver to test against

        Returns:
            bool: If the solver is legal
        """
        if any(isinstance(solver, solvertype) for solvertype in self.disabled_solver):
            return False
        return True
    
    @property
    def all_solvers(self) -> list[Solver]:
        return list([solver for solver in self.solvers.values() if not isinstance(solver, EigSolver)])
    
    @property
    def all_eig_solvers(self) -> list[EigSolver]:
        return list([solver for solver in self.solvers.values() if isinstance(solver, EigSolver)])
    

    def _try_solver(self, solver_type: EMSolver) -> Solver:
        """Try to use the selected solver or else find another one that is working.

        Args:
            solver_type (EMSolver): The solver type to try

        Raises:
            RuntimeError: Error if no valid solver is found.

        Returns:
            Solver: The working solver.
        """
        solver = self.solvers[solver_type]
        if self._legal_solver(solver):
            return solver  # type: ignore
        for alternative in self.all_solvers:
            if self._legal_solver(alternative):
                logger.debug(f'Falling back on legal solver: {alternative}')
                return alternative
        raise RuntimeError(f'No legal solver could be found. The following are disabled: {self.disabled_solver}')
    
    def duplicate(self) -> SolveRoutine:
        """Creates a copy of this SolveRoutine class object.

        Returns:
            SolveRoutine: The copied version
        """
        new_routine = self.__class__()
        new_routine.parallel = self.parallel
        new_routine.smart_search = self.smart_search
        new_routine.forced_solver = self.forced_solver
        for tpe, solver in self.solvers.items():
            new_routine.solvers[tpe] = solver.duplicate()
        return new_routine
    
    def set_solver(self, *solvers: EMSolver | EigSolver | Solver) -> None:
        """Set a given Solver class instance as the main solver. Solvers will be checked on validity for the given problem.

        Args:
            solver (EMSolver | Solver): The solver objects
        """
        for solver in solvers:
            if isinstance(solver, EMSolver):
                self.forced_solver.append(solver.get_solver())
            else:
                self.forced_solver.append(solver)
    
    def disable(self, *solvers: EMSolver) -> None:
        """Disable a given Solver class instance as the main solver. Solvers will be checked on validity for the given problem.

        Args:
            solver (EMSolver): The solver objects
        """
        for solver in solvers:
            if isinstance(solver, EMSolver):
                self.disabled_solver.append(solver.get_solver().__class__)
            else:
                self.disabled_solver.append(solver.__class__)

    def _configure_routine(self, 
                  parallel: Literal['SI','MT','MP'] = 'SI', 
                  smart_search: bool = False) -> SolveRoutine:
        """Configure the solver with the given settings

        Args:
            parallel (Literal['SI','MT','MP'], optional): 
                The solver parallism, Defaults to 'SI'.
                    - "SI" = Single threaded
                    - "MT" = Multi threaded
                    - "MP" = Multi-processing,
            smart_search (bool, optional): Wether to use smart-search solvers for eigenmode problems. Defaults to False.

        Returns:
            SolveRoutine: The same SolveRoutine object.
        """
        self.parallel = parallel
        self.smart_search = smart_search
        return self

    def configure(self,
                    pivoting_threshold: float | None = None) -> None:
        """Sets general user configurations for all solvers.

        Args:
            pivoting_threshold (float | None, optional): 
                The diagonal pivoting threshold used in direct solvers. Standard values are 0.001.
                In simulations with a very low surface impedance (such as with copper walls) a much
                lower pivoting threshold is desired.
        """
        for solver in self.solvers.values():
            if isinstance(solver, Solver):
                solver.set_options(pivoting_threshold=pivoting_threshold)
        
    def reset(self) -> None:
        """Reset all solver states"""
        for solver in self.solvers.values():
            solver.reset()
        self.sorter.reset()
        self.parallel = 'SI'
        self.smart_search = False
        self.forced_solver = []
        self.disabled_solver = []

    def _get_solver(self, A: csr_matrix, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: Returns the direct solver

        """
        for solver in self.forced_solver:
            if not self._legal_solver(solver):
                continue
            if isinstance(solver, Solver):
                return solver
        return self.pick_solver(A,b)
        
    
    def pick_solver(self, A: csr_matrix, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: Returns the direct solver

        """
        return self._try_solver(EMSolver.SUPERLU)
    
    def _get_eig_solver(self, A: csr_matrix, b: csr_matrix, direct: bool | None = None) -> EigSolver:
        """Returns the relevant eigenmode Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for
            direct (bool): If the direct solver should be used.

        Returns:
            Solver: Returns the solver object

        """
        for solver in self.forced_solver:
            if isinstance(solver, EigSolver):
                return solver  # type: ignore
        if direct or A.shape[0] < 1000:
            return self.solvers[EMSolver.LAPACK] # type: ignore
        else:
            return self.solvers[EMSolver.SMART_ARPACK] # type: ignore
            
    def _get_eig_solver_bma(self, A: csr_matrix, b: csr_matrix, direct: bool | None = None) -> EigSolver:
        """Returns the relevant eigenmode Solver object given a certain matrix and source vector

        This is the default implementation for the SolveRoutine Class.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for
            direct (bool): If the direct solver should be used.

        Returns:
            Solver: Returns the solver object

        """
        for solver in self.forced_solver:
            if isinstance(solver, EigSolver):
                return solver
        
        if direct or A.shape[0] < 1000:
            return self.solvers[EMSolver.LAPACK]  # type: ignore
        else:
            return self.solvers[EMSolver.ARPACK]  # type: ignore
    
    def solve(self, A: csr_matrix | csr_matrix, 
              b: np.ndarray, 
              solve_ids: np.ndarray,
              reuse: bool = False,
              id: int = -1) -> tuple[np.ndarray, SolveReport]:
        """ Solve the system of equations defined by Ax=b for x.

        Solve is the main function call to solve a linear system of equations defined by Ax=b.
        The solve routine will go through the required steps defined in the routine to tackle the problme.

        Args:
            A (np.ndarray | csr_matrix | csr_matrix): The (Sparse) matrix
            b (np.ndarray): The source vector
            solve_ids (np.ndarray): A vector of ids for which to solve the problem. For EM problems this
            implies all non-PEC degrees of freedom.
            reuse (bool): Whether to reuse the existing factorization if it exists.

        Returns:
            np.ndarray: The resultant solution.
        """
        solver: Solver = self._get_solver(A, b)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        A = A.tocsc()
        
        Asel = A[np.ix_(solve_ids, solve_ids)]
        bsel = b[solve_ids]
        nnz = Asel.nnz

        logger.debug(f'[ID={id}] Removed {NF-NS} prescribed DOFs ({NS:,} left, {nnz:,}≠0)')

        if solver.real_only:
            logger.debug(f'[ID={id}] Converting to real matrix')
            Asel, bsel = complex_to_real_block(Asel, bsel)

        # SORT
        sorter = 'None'
        if solver.req_sorter and self.use_sorter:
            sorter = str(self.sorter)
            Asorted, bsorted = self.sorter.sort(Asel, bsel, reuse_sorting=reuse)
        else:
            Asorted, bsorted = Asel, bsel
        
        # Preconditioner
        precon = 'None'
        if self.use_preconditioner:
            if not solver.own_preconditioner:
                self.precon.init(Asorted, bsorted)
                precon = str(self.precon)

        start = time.time()
        
        x_solved, report = solver.solve(Asorted, bsorted, self.precon, reuse_factorization=reuse, id=id)
        end = time.time()
        simtime = end-start
        logger.info(f'[ID={id}] Elapsed time taken: {simtime:.3f} seconds')
        logger.debug(f'[ID={id}] O(N²) performance = {(NS**2)/((end-start+1e-6)*1e6):.3f} MDoF/s')
        
        if self.use_sorter and solver.req_sorter:
            x = self.sorter.unsort(x_solved)
        else:
            x = x_solved
        
        if solver.real_only:
            logger.debug(f'[ID={id}] Converting back to complex matrix')
            x = real_to_complex_block(x)

        solution = np.zeros((NF,), dtype=np.complex128)
        
        solution[solve_ids] = x

        logger.debug(f'[ID={id}] Solver complete!')
        report.jobid = id
        report.sorter = str(sorter)
        report.simtime = simtime
        report.nnz = A.nnz
        report.ndof = b.shape[0]
        report.nnz_solve = Asorted.nnz
        report.ndof_solve = bsorted.shape[0]
        report.precon = precon

        return solution, report
    
    def eig_boundary(self, 
            A: csr_matrix | csr_matrix, 
            B: np.ndarray, 
            solve_ids: np.ndarray,
            nmodes: int = 6,
            direct: bool | None = None,
            target_k0: float = 0.0,
            which: str = 'LM', 
            sign: float=-1) -> tuple[np.ndarray, np.ndarray, SolveReport]:
        """ Find the eigenmodes for the system Ax = λBx for a boundary mode problem

        For generalized eigenvalue problems of boundary mode analysis studies, the equation is: Ae = -β²Be

        Args:
            A (csr_matrix): The Stiffness matrix
            B (csr_matrix): The mass matrix
            solve_ids (np.ndarray): The free nodes (non PEC)
            nmodes (int): The number of modes to solve for. Defaults to 6
            direct (bool): If the direct solver should be used (always). Defaults to False
            target_k0 (float): The k0 value to search around
            which (str): The search method. Defaults to 'LM' (Largest Magnitude)
            sign (float): The sign of the eigenvalue expression. Defaults to -1

        Returns:
            np.ndarray: The eigen values
            np.ndarray: The eigen vectors
            SolveReport: The solution report
        """
        solver = self._get_eig_solver_bma(A, B, direct=direct)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        logger.debug(f'Removing {NF-NS} prescribed DOFs ({NS} left)')

        Asel = A[np.ix_(solve_ids, solve_ids)]
        Bsel = B[np.ix_(solve_ids, solve_ids)]
        
        start = time.time()
        eigen_values, eigen_modes = solver.eig(Asel, Bsel, nmodes, target_k0, which, sign=sign)
        end = time.time()

        simtime = end-start
        return eigen_values, eigen_modes, SolveReport(ndof=A.shape[0], nnz=A.nnz, ndof_solve=Asel.shape[0], nnz_solve=Asel.nnz, simtime=simtime, solver=str(solver), sorter='None', precon='None')

    def eig(self, 
            A: csr_matrix | csr_matrix, 
            B: np.ndarray, 
            solve_ids: np.ndarray,
            nmodes: int = 6,
            direct: bool | None = None,
            target_f0: float = 0.0,
            which: str = 'LM') -> tuple[np.ndarray, np.ndarray, SolveReport]:
        """
        Find the eigenmodes for the system Ax = λBx for a boundary mode problem
        
        Args:
            A (csr_matrix): The Stiffness matrix
            B (csr_matrix): The mass matrix
            solve_ids (np.ndarray): The free nodes (non PEC)
            nmodes (int): The number of modes to solve for. Defaults to 6
            direct (bool): If the direct solver should be used (always). Defaults to False
            target_k0 (float): The k0 value to search around
            which (str): The search method. Defaults to 'LM' (Largest Magnitude)
            sign (float): The sign of the eigenvalue expression. Defaults to -1\
        Returns:
            np.ndarray: The resultant solution.
        """
        solver = self._get_eig_solver(A, B, direct=direct)

        NF = A.shape[0]
        NS = solve_ids.shape[0]

        logger.debug(f'Removing {NF-NS} prescribed DOFs ({NS} left)')

        Asel = A[np.ix_(solve_ids, solve_ids)]
        Bsel = B[np.ix_(solve_ids, solve_ids)]
        
        start = time.time()
        eigen_values, eigen_modes = solver.eig(Asel, Bsel, nmodes, target_f0, which, sign=1.0)
        end = time.time()
        simtime = end-start

        Nsols = eigen_modes.shape[1]
        sols = np.zeros((NF, Nsols), dtype=np.complex128)
        for i in range(Nsols):
            sols[solve_ids,i] = eigen_modes[:,i]

        return eigen_values, sols, SolveReport(ndof=A.shape[0], nnz=A.nnz, ndof_solve=Asel.shape[0], nnz_solve=Asel.nnz, simtime=simtime, solver=str(solver), sorter='None', precon='None')


class AutomaticRoutine(SolveRoutine):
    """ Defines the Automatic Routine for EMerge.
    """
        
    def pick_solver(self, A: np.ndarray, b: np.ndarray) -> Solver:
        """Returns the relevant Solver object given a certain matrix and source vector

        The current implementation only looks at matrix size to select the best solver. Matrices 
        with a large size will use iterative solvers while smaller sizes will use either Pardiso
        for medium sized problems or SPSolve for small ones.
        
        Args:
            A (np.ndarray): The Matrix to solve for
            b (np.ndarray): the vector to solve for

        Returns:
            Solver: A solver object appropriate for solving the problem.

        """
        N = b.shape[0]
        if N < 10_000:
            return self._try_solver(EMSolver.SUPERLU)
        if self.parallel=='SI':
            if _PARDISO_AVAILABLE:
                return self._try_solver(EMSolver.PARDISO)
            elif _UMFPACK_AVAILABLE:
                return self._try_solver(EMSolver.UMFPACK)
            else:
                return self._try_solver(EMSolver.SUPERLU)
        elif self.parallel=='MP':
            if _UMFPACK_AVAILABLE:
                return self._try_solver(EMSolver.UMFPACK)
            else:
                return self._try_solver(EMSolver.SUPERLU)
        elif self.parallel=='MT':
            return self._try_solver(EMSolver.SUPERLU)
        return self._try_solver(EMSolver.SUPERLU)
    


############################################################
#                    DEFAULT DEFINITION                   #
############################################################

DEFAULT_ROUTINE = AutomaticRoutine()


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the TERMS of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.


import numpy as np
from numba import njit, types, i8, c16
from scipy.sparse import csr_matrix


############################################################
#                      NUMBA COMPILED                     #
############################################################

@njit(types.Tuple((i8[:], i8[:], c16[:], c16[:]))(i8[:,:], i8[:,:], i8[:,:], i8[:,:], i8), cache=True, nogil=True)
def _fill_periodic_matrix(tris: np.ndarray, edges: np.ndarray, tri_to_field: np.ndarray, edge_to_field: np.ndarray, Nfield: int) -> tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Generates sparse matrix row, column ids and ones plus the matrix diagonal.

    Args:
        tris (: np.ndarray): The triangle ids
        edges (: np.ndarray): The edge ids
        tri_to_field (: np.ndarray): The triangle to field index mapping
        edge_to_field (: np.ndarray): The edge to field index mapping
        Nfield (int): The number of field points

    Returns:
        np.ndarray: The row ids
        np.ndarray: The column ids
        np.ndarray: The data
        np.ndarray: The diagonal array
    """

    # NUMBERS
    N = tris.shape[1] + edges.shape[1]
    NT = tris.shape[1]
    
    DIAGONAL = np.ones((Nfield,), dtype=np.complex128)
    ROWS = np.zeros((N*2,), dtype=np.int64)
    COLS = np.zeros((N*2,), dtype=np.int64)
    TERMS = np.zeros((N*2,), dtype=np.complex128)
    
    i = 0
    for it in range(NT):
        t1 = tris[0,it]
        t2 = tris[1,it]
        f11 = tri_to_field[3, t1]
        f21 = tri_to_field[7, t1]
        f12 = tri_to_field[3, t2]
        f22 = tri_to_field[7, t2]
        DIAGONAL[f12] = 0.
        DIAGONAL[f22] = 0.
        ROWS[i] = f12
        ROWS[i+1] = f22
        COLS[i] = f11
        COLS[i+1] = f21
        TERMS[i] = 1.0
        TERMS[i+1] = 1.0
        i += 2
    NE = edges.shape[1]
    for ie in range(NE):
        e1 = edges[0,ie]
        e2 = edges[1,ie]
        f11 = edge_to_field[0, e1]
        f21 = edge_to_field[1, e1]
        f12 = edge_to_field[0, e2]
        f22 = edge_to_field[1, e2]
        DIAGONAL[f12] = 0.
        DIAGONAL[f22] = 0.
        ROWS[i] = f12
        ROWS[i+1] = f22
        COLS[i] = f11
        COLS[i+1] = f21
        TERMS[i] = 1.0
        TERMS[i+1] = 1.0
        i += 2
    ROWS = ROWS[:i]
    COLS = COLS[:i]
    TERMS = TERMS[:i]
    return ROWS, COLS, TERMS, DIAGONAL


############################################################
#                     PYTHON INTERFACE                    #
############################################################

def gen_periodic_matrix(tris: np.ndarray, 
                        edges: np.ndarray, 
                        tri_to_field: np.ndarray, 
                        edge_to_field: np.ndarray, 
                        linked_tris: dict[int, int], 
                        linked_edges: dict[int, int], 
                        Nfield: int, 
                        phi: complex) -> tuple[csr_matrix, np.ndarray]:
    """This function constructs the periodic boundary matrix

    Args:
        tris (np.ndarray): _description_
        edges (np.ndarray): _description_
        tri_to_field (np.ndarray): _description_
        edge_to_field (np.ndarray): _description_
        linked_tris (dict[int, int]): _description_
        linked_edges (dict[int, int]): _description_
        Nfield (int): _description_
        phi (complex): _description_

    Returns:
        tuple[csr_matrix, np.ndarray]: _description_
    """

    tris_array = np.array([(tri, linked_tris[tri]) for tri in tris]).T
    edges_array = np.array([(edge, linked_edges[edge]) for edge in edges]).T
    ROWS, COLS, TERMS, diagonal = _fill_periodic_matrix(tris_array, edges_array, tri_to_field, edge_to_field, Nfield)
    matrix = csr_matrix((TERMS, (ROWS, COLS)), [Nfield, Nfield], dtype=np.complex128)
    matrix.data.fill(phi)
    matrix.setdiag(diagonal)
    
    return matrix, ROWS


import emerge as em

"""DEMO 7: Periodic Cells

Since version 0.2 of EMerge, there is a good support for periodic environemnts.
The setup requires some manual steps. In this demonstration we will look at seting up a rectangular waveguide
array in a flat hexagonal periodic Tiling.
+-----------------+-----------------+-----------------+
|  +-----------+  |  +-----------+  |  +-----------+  |
|  |           |  |  |           |  |  |           |  |
|  +-----------+  |  +-----------+  |  +-----------+  |
+--------+--------+--------+--------+--------+--------+
         |  +-----------+  |  +-----------+  |
         |  |           |  |  |           |  |
         |  +-----------+  |  +-----------+  |
+--------+--------+--------+--------+--------+--------+
|  +-----------+  |  +-----------+  |  +-----------+  |
|  |           |  |  |           |  |  |           |  |
|  +-----------+  |  +-----------+  |  +-----------+  |
+--------+--------+--------+--------+--------+--------+
"""
mm = 0.001
a = 106*mm
b = 30*mm
H = 70*mm
wga = 70*mm
wgb = 18*mm
fl = 25*mm

# We start again by defining our simulation model
model = em.Simulation3D('Periodic', loglevel='DEBUG')

# Next we will create a PeriodicCell class (in our case a hexagonal cell). This class
# is simply meant to simplify our lives and improve the simulation setup flow.
# To define our hexagonal cell we have to specify the three points that make up our hexagonal grid. For a normal
# hexagon the points would be the following

#             _____
#            /     \
#           /       \
#     ,----(         )----.
#    /      \       /      \
#   /       (1)____/        \
#   \        /     \        /
#    \      /       \      /
#     )---(2)        )----(
#    /      \       /      \
#   /        \_____/        \
#   \       (3)    \        /
#    \      /       \      /
#     `----(         )----'
#           \       /
#            \_____/

# In the case of our rectangular waveguide array we will use the following;
#(1)-------+--------+--------+--------+--------+--------+
# |  +-----------+  |  +-----------+  |  +-----------+  |
# |  |           |  |  |           |  |  |           |  |
# |  +-----------+  |  +-----------+  |  +-----------+  |
#(2)------(3)-------+--------+--------+--------+--------+

periodic_cell = em.HexCell((-a/2, b/2, 0), (-a/2, -b/2, 0), (0, -b/2, 0))

# To make sure that we can run a periodic simulation we must tell the simulation that
# it has to copy the meshing on each face that is duplcated. We can simply pass our periodic
# cell to our model using the set_periodic_cell() method.

model.set_periodic_cell(periodic_cell)

# We can easily use our periodic cell to construct volumes with the appropriate faces. We simply call the volume method
# to construct a cell region from z=0 to z=H

model['box'] = periodic_cell.volume(0, H)

# We also create a waveguide foor the feed
model['wg'] = em.geo.Box(wga,wgb,fl, (-wga/2, -wgb/2,-fl) )

# Next we define our geometry as usual
# Beause we stored our geometry in our model object using the get and set-item notation. We don't have to pass the items anymore.
model.commit_geometry()

model.mw.set_frequency_range(2.8e9, 3.3e9, 5)
model.mw.set_resolution(0.1)

# Then we create our mesh and view the result
model.generate_mesh()
model.view()

# Now lets define our boundary conditions
# First the waveguide port
wg = model['wg']
wgbc = model.mw.bc.RectangularWaveguide(wg.face('bottom'), 1)

# And then the absorbing boundary at the top
abc = model.mw.bc.AbsorbingBoundary(model['box'].face('back'))

# We can use the set_scanangle method to set the appropriate phases for the boundary. The scan angle is defined as following
# kx = sin(θ)·cos(ϕ)
# ky = sin(θ)·sin(ϕ)
# kx = cos(θ)ϕ
# The arguments of the function are θ,ϕ in degrees.
periodic_cell.set_scanangle(30,45)

# And at last we run our simulation and view the results.
data = model.mw.frequency_domain()

model.display.add_object(wg)
model.display.add_object(model['box'])
model.display.add_surf(*data.field[0].cutplane(3*mm, y=0).scalar('Ey','real'))
model.display.add_surf(*data.field[0].cutplane(3*mm, x=0).scalar('Ey','real'))
model.display.show()




class PVDisplaySettings:

    def __init__(self):
        self.draw_xplane: bool = True
        self.draw_yplane: bool = True
        self.draw_zplane: bool = True
        self.draw_xax: bool = True
        self.draw_yax: bool = True
        self.draw_zax: bool = True
        self.plane_ratio: float = 0.5
        self.plane_opacity: float = 0.00
        self.plane_edge_width: float = 1.0
        self.axis_line_width: float = 1.5
        self.show_xgrid: bool = False
        self.show_ygrid: bool = False
        self.show_zgrid: bool = True
        self.grid_line_width: float = 1.0
        self.add_light: bool = False
        self.light_angle: tuple[float, float] = (20., -20.)
        self.cast_shadows: bool = True
        self.background_bottom: str = "#c0d2e8"
        self.background_top: str = "#ffffff"
        self.grid_line_color: str = "#8e8e8e"



"""A Python based FEM solver.
Copyright (C) 2025 Robert Fennis

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, see
<https://www.gnu.org/licenses/>.

"""
import os


############################################################
#               HANDLE ENVIRONMENT VARIABLES              #
############################################################

NTHREADS = "1"
os.environ["EMERGE_STD_LOGLEVEL"] = os.getenv("EMERGE_STD_LOGLEVEL", default="INFO")
os.environ["EMERGE_FILE_LOGLEVEL"] = os.getenv("EMERGE_FILE_LOGLEVEL", default="DEBUG")
os.environ["OMP_NUM_THREADS"] = os.getenv("OMP_NUM_THREADS", default="4")
os.environ["MKL_NUM_THREADS"] = os.getenv("MKL_NUM_THREADS", default="4")
os.environ["OPENBLAS_NUM_THREADS"] = NTHREADS
os.environ["VECLIB_MAXIMUM_THREADS"] = NTHREADS
os.environ["NUMEXPR_NUM_THREADS"] = NTHREADS


############################################################
#                      IMPORT MODULES                     #
############################################################
from ._emerge import _cache_check
from ._emerge.logsettings import LOG_CONTROLLER
from loguru import logger

LOG_CONTROLLER.set_default()
logger.debug('Importing modules')

from ._emerge.simmodel import Simulation3D
from ._emerge.material import Material
from ._emerge import bc
from ._emerge.solver import SolverBicgstab, SolverGMRES, SolveRoutine, ReverseCuthillMckee, Sorter, SolverPardiso, SolverUMFPACK, SolverSuperLU, EMSolver
from ._emerge.cs import CoordinateSystem, CS, GCS, Plane, Axis, XAX, YAX, ZAX, XYPLANE, XZPLANE, YZPLANE, YXPLANE, ZXPLANE, ZYPLANE
from ._emerge.coord import Line
from ._emerge import geo
from ._emerge.selection import Selection, FaceSelection, DomainSelection, EdgeSelection
from ._emerge.mth.common_functions import norm, coax_rout, coax_rin
from ._emerge.physics.microwave.sc import stratton_chu
from ._emerge.periodic import RectCell, HexCell
from ._emerge.mesher import Algorithm2D, Algorithm3D
from . import lib
from ._emerge.howto import _HowtoClass
howto = _HowtoClass()
logger.debug('Importing complete!')


"""
╔══════════════════════════════════════════════════════════════════════╗
║  library.py – Material-property database                             ║
║                                                                      ║
║  This file is released under CC0 1.0 – Public-Domain Dedication.     ║
║  You may use, copy, modify, and distribute it without restriction.   ║
║                                                                      ║
║  All OTHER files in this repository are proprietary. See LICENSE     ║
║  at the project root for full terms, conditions, and disclaimers.    ║
║                                                                      ║
║  DISCLAIMER: Data herein are provided “as is”, with no warranties.   ║
║  Verify critical values independently before use.                    ║
╚══════════════════════════════════════════════════════════════════════╝
"""
from ._emerge.material import Material, AIR, COPPER
from ._emerge.const import C0, Z0, PI, EPS0, MU0

EISO: float = (Z0/(2*PI))**0.5
EOMNI = (3*Z0/(4*PI))**0.5

##MATERIALS 
VACUUM = Material(color="#2d8cd5", opacity=0.05)

############################################################
#                         METALS                         #
############################################################

GREY = "#bfbfbf"
MET_ALUMINUM         = Material(cond=3.77e7,  color=GREY, opacity=0.5)
MET_CARBON           = Material(cond=3.33e4,  color=GREY, opacity=0.5)
MET_CHROMIUM         = Material(cond=5.56e6,  color=GREY, opacity=0.5)
MET_COPPER           = Material(cond=5.8e7,   color="#62290c", opacity=0.5)
MET_GOLD             = Material(cond=4.10e7,  color=GREY, opacity=0.5)
MET_INDIUM           = Material(cond=6.44e6,  color=GREY, opacity=0.5)
MET_IRIDIUM          = Material(cond=2.13e7,  color=GREY, opacity=0.5)
MET_IRON             = Material(cond=1.04e7,  color=GREY, opacity=0.5)
MET_LEAD             = Material(cond=4.84e6,  color=GREY, opacity=0.5)
MET_MAGNESIUM        = Material(cond=2.38e7,  color=GREY, opacity=0.5)
MET_NICKEL           = Material(cond=1.14e7,  color=GREY, opacity=0.5)
MET_NICHROME         = Material(cond=9.09e5,  color=GREY, opacity=0.5)
MET_PALLADIUM        = Material(cond=9.42e6,  color=GREY, opacity=0.5)
MET_PLATINUM         = Material(cond=9.42e6,  color=GREY, opacity=0.5)
MET_RHODIUM          = Material(cond=2.22e7,  color=GREY, opacity=0.5)
MET_SILVER           = Material(cond=6.29e7,  color=GREY, opacity=0.5)
MET_TANTALUM         = Material(cond=6.44e6,  color=GREY, opacity=0.5)
MET_TANTALUM_NITRIDE = Material(cond=3.97e5,  color=GREY, opacity=0.5)
MET_TIN              = Material(cond=8.66e6,  color=GREY, opacity=0.5)
MET_TITANIUM         = Material(cond=1.82e6,  color=GREY, opacity=0.5)
MET_TUNGSTEN         = Material(cond=1.79e7,  color=GREY, opacity=0.5)
MET_ZINC             = Material(cond=1.76e7,  color=GREY, opacity=0.5)
MET_ZIRCONIUM        = Material(cond=2.44e7,  color=GREY, opacity=0.5)


############################################################
#                       SEMICONDUCTORS                     #
############################################################

SEMI_SILICON       = Material(er=11.7,  tand=0.005,     color="#b4b4b4", opacity=0.5) #  Crystalline Si
SEMI_SILICON_N     = Material(er=7.5,   tand=0.0003,    color="#a0a0a0", opacity=0.5) #  Silicon Nitride (Si₃N₄)
SEMI_SILICON_OXIDE = Material(er=3.9,   tand=0.0001,    color="#e0e0e0", opacity=0.5) #  Silicon Dioxide (SiO₂)
SEMI_GERMANIUM     = Material(er=16.0,  tand=0.001,     color="#787878", opacity=0.5)
SEMI_GAAS          = Material(er=13.1,  tand=0.0016,    color="#aa8888", opacity=0.5) #  Gallium Arsenide
SEMI_GA_N          = Material(er=8.9,   tand=0.002,     color="#8888cc", opacity=0.5) #  Gallium Nitride
SEMI_INP           = Material(er=12.5,  tand=0.0015,    color="#cc99aa", opacity=0.5) #  Indium Phosphide
SEMI_ALN           = Material(er=8.6,   tand=0.0003,    color="#ccccee", opacity=0.5) #  Aluminum Nitride
SEMI_AL2O3         = Material(er=9.8,   tand=0.0002,    color="#eaeaea", opacity=0.5) #  Alumina
SEMI_SAPPHIRE      = Material(er=9.4,   tand=0.0001,    color="#ddddff", opacity=0.5)
SEMI_DIAMOND       = Material(er=5.5,   tand=0.00005,   color="#cceeff", opacity=0.5) #  Synthetic CVD diamond
SEMI_HBN           = Material(er=4.0,   tand=0.0001,    color="#eeeeff", opacity=0.5) #  Hexagonal Boron Nitride
SEMI_SIOXNY        = Material(er=5.0,   tand=0.002,     color="#ddddee", opacity=0.5) #  Silicon Oxynitride (SiOxNy)

############################################################
#                          LIQUIDS                         #
############################################################

LIQ_WATER     = Material(er=80.1,    cond=0.0,  color="#0080ff", opacity=0.3)
LIQ_FERRITE   = Material(er=12.0,    ur=2000, tand=0.02, color="#994d4d", opacity=0.3)

############################################################
#                        DIELECTRICS                       #
############################################################

#  TRADEMARKS: All product names (e.g. “DUROID”) are the property of their
#  respective owners.  Use of them here does not imply any affiliation with or
#  endorsement by those owners.

DIEL_PTFE = Material(er=2.1, tand=0.0002, color="#21912b", opacity=0.3)
DIEL_POLYIMIDE = Material(er=3.4, tand=0.02, color="#b8b8b8")
DIEL_CERAMIC = Material(er=6.0, tand=0.001, color="#efead1")
DIEL_AD10 = Material(er=10.2, tand=0.0078, color="#21912b", opacity=0.3)
DIEL_AD1000 = Material(er=10.2, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_AD250 = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_AD250_PIM = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_AD250A = Material(er=2.50, tand=0.0015, color="#21912b", opacity=0.3)
DIEL_AD250C = Material(er=2.50, tand=0.0014, color="#21912b", opacity=0.3)
DIEL_AD255 = Material(er=2.55, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_AD255A = Material(er=2.55, tand=0.0015, color="#21912b", opacity=0.3)
DIEL_AD255C = Material(er=2.55, tand=0.0014, color="#21912b", opacity=0.3)
DIEL_AD260A = Material(er=2.60, tand=0.0017, color="#21912b", opacity=0.3)
DIEL_AD270 = Material(er=2.7, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_AD300 = Material(er=3, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD300_PIM = Material(er=3, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD300A = Material(er=3.00, tand=0.002, color="#21912b", opacity=0.3)
DIEL_AD300C = Material(er=2.97, tand=0.002, color="#21912b", opacity=0.3)
DIEL_AD320 = Material(er=3.2, tand=0.0038, color="#21912b", opacity=0.3)
DIEL_AD320_PIM = Material(er=3.2, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD320A = Material(er=3.20, tand=0.0032, color="#21912b", opacity=0.3)
DIEL_AD350 = Material(er=3.5, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD350_PIM = Material(er=3.5, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD350A = Material(er=3.50, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD410 = Material(er=4.1, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD430 = Material(er=4.3, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD450 = Material(er=4.5, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_AD450A = Material(er=4.5, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_AD5 = Material(er=5.1, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AD600 = Material(er=5.90, tand=0.003, color="#21912b", opacity=0.3)
DIEL_AR1000 = Material(er=9.8, tand=0.003, color="#21912b", opacity=0.3)
DIEL_CER_10 = Material(er=10.00, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_CLTE = Material(er=2.96, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_CLTE_AT = Material(er=3.00, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_CLTE_LC = Material(er=2.94, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_CLTE_XT = Material(er=2.94, tand=0.0012, color="#21912b", opacity=0.3)
DIEL_COMCLAD_HF_ER2 = Material(er=2, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_COMCLAD_HF_ER3 = Material(er=3, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_COMCLAD_HF_ER4 = Material(er=4, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_COMCLAD_HF_ER5 = Material(er=5, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_COMCLAD_HF_ER6 = Material(er=6, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_COPPER_CLAD_ULTEM = Material(er=3.05, tand=0.003, color="#21912b", opacity=0.3)
DIEL_CUCLAD_217LX = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_CUCLAD_233LX = Material(er=2.33, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_CUCLAD_250GT = Material(er=2.5, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_CUCLAD_250GX = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_CUFLON = Material(er=2.05, tand=0.00045, color="#21912b", opacity=0.3)
DIEL_DICLAD_522 = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_DICLAD_527 = Material(er=2.4, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_DICLAD_870 = Material(er=2.33, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_DICLAD_880 = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_DICLAD_880_PIM = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_GETEK = Material(er=3.5, tand=0.01, color="#21912b", opacity=0.3)
DIEL_GETEK = Material(er=3.8, tand=0.01, color="#21912b", opacity=0.3)
DIEL_IS6802_80 = Material(er=2.80, tand=0.003, color="#21912b", opacity=0.3)
DIEL_IS6803_00 = Material(er=3.00, tand=0.003, color="#21912b", opacity=0.3)
DIEL_IS6803_20 = Material(er=3.20, tand=0.003, color="#21912b", opacity=0.3)
DIEL_IS6803_33 = Material(er=3.33, tand=0.003, color="#21912b", opacity=0.3)
DIEL_IS6803_38 = Material(er=3.38, tand=0.0032, color="#21912b", opacity=0.3)
DIEL_IS6803_45 = Material(er=3.45, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_ISOCLAD_917 = Material(er=2.17, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_ISOCLAD_933 = Material(er=2.33, tand=0.0016, color="#21912b", opacity=0.3)
DIEL_ISOLA_I_TERA_MT = Material(er=3.45, tand=0.0030,      color="#3c9747")
DIEL_ISOLA_NELCO_4000_13 = Material(er=3.77, tand=0.008,  color="#3c9747")
DIEL_MAT_25N = Material(er=3.38, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_MAT25FR = Material(er=3.58, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_MEGTRON6R5775 = Material(er=3.61, tand=0.004, color="#21912b", opacity=0.3)
DIEL_MERCURYWAVE_9350 = Material(er=3.5, tand=0.004, color="#21912b", opacity=0.3)
DIEL_MULTICLAD_HF = Material(er=3.7, tand=0.0045, color="#21912b", opacity=0.3)
DIEL_N_8000 = Material(er=3.5, tand=0.011, color="#21912b", opacity=0.3)
DIEL_N4350_13RF = Material(er=3.5, tand=0.0065, color="#21912b", opacity=0.3)
DIEL_N4380_13RF = Material(er=3.8, tand=0.007, color="#21912b", opacity=0.3)
DIEL_N8000Q = Material(er=3.2, tand=0.006, color="#21912b", opacity=0.3)
DIEL_N9300_13RF = Material(er=3, tand=0.004, color="#21912b", opacity=0.3)
DIEL_N9320_13RF = Material(er=3.2, tand=0.0045, color="#21912b", opacity=0.3)
DIEL_N9338_13RF = Material(er=3.38, tand=0.0046, color="#21912b", opacity=0.3)
DIEL_N9350_13RF = Material(er=3.48, tand=0.0055, color="#21912b", opacity=0.3)
DIEL_NH9294 = Material(er=2.94, tand=0.0022, color="#21912b", opacity=0.3)
DIEL_NH9300 = Material(er=3.00, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_NH9320 = Material(er=3.20, tand=0.0024, color="#21912b", opacity=0.3)
DIEL_NH9338 = Material(er=3.38, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_NH9348 = Material(er=3.48, tand=0.003, color="#21912b", opacity=0.3)
DIEL_NH9350 = Material(er=3.50, tand=0.003, color="#21912b", opacity=0.3)
DIEL_NH9410 = Material(er=4.10, tand=0.003, color="#21912b", opacity=0.3)
DIEL_NH9450 = Material(er=4.50, tand=0.003, color="#21912b", opacity=0.3)
DIEL_NORCLAD = Material(er=2.55, tand=0.0011, color="#21912b", opacity=0.3)
DIEL_NX9240 = Material(er=2.40, tand=0.0016, color="#21912b", opacity=0.3)
DIEL_NX9245 = Material(er=2.45, tand=0.0016, color="#21912b", opacity=0.3)
DIEL_NX9250 = Material(er=2.50, tand=0.0017, color="#21912b", opacity=0.3)
DIEL_NX9255 = Material(er=2.55, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_NX9260 = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_NX9270 = Material(er=2.70, tand=0.002, color="#21912b", opacity=0.3)
DIEL_NX9294 = Material(er=2.94, tand=0.0022, color="#21912b", opacity=0.3)
DIEL_NX9300 = Material(er=3.00, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_NX9320 = Material(er=3.20, tand=0.0024, color="#21912b", opacity=0.3)
DIEL_NY9208 = Material(er=2.08, tand=0.0006, color="#21912b", opacity=0.3)
DIEL_NY9217 = Material(er=2.17, tand=0.0008, color="#21912b", opacity=0.3)
DIEL_NY9220 = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_NY9233 = Material(er=2.33, tand=0.0011, color="#21912b", opacity=0.3)
DIEL_POLYGUIDE = Material(er=2.320, tand=0.0005, color="#21912b", opacity=0.3)
DIEL_RF_30 = Material(er=3.00, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_RF_301 = Material(er=2.97, tand=0.0018, color="#21912b", opacity=0.3)
DIEL_RF_35 = Material(er=3.50, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_RF_35A2 = Material(er=3.50, tand=0.0015, color="#21912b", opacity=0.3)
DIEL_RF_35P = Material(er=3.50, tand=0.0034, color="#21912b", opacity=0.3)
DIEL_RF_35TC = Material(er=3.50, tand=0.0011, color="#21912b", opacity=0.3)
DIEL_RF_41 = Material(er=4.10, tand=0.0038, color="#21912b", opacity=0.3)
DIEL_RF_43 = Material(er=4.30, tand=0.0033, color="#21912b", opacity=0.3)
DIEL_RF_45 = Material(er=4.50, tand=0.0037, color="#21912b", opacity=0.3)
DIEL_RF_60A = Material(er=6.15, tand=0.0038, color="#21912b", opacity=0.3)
DIEL_RO3003 = Material(er=3.00, tand=0.0011, color="#21912b", opacity=0.3)
DIEL_RO3006 = Material(er=6.15, tand=0.002, color="#21912b", opacity=0.3)
DIEL_RO3010 = Material(er=10.2, tand=0.0022, color="#21912b", opacity=0.3)
DIEL_RO3035 = Material(er=3.50, tand=0.0017, color="#21912b", opacity=0.3)
DIEL_RO3203 = Material(er=3.02, tand=0.0016, color="#21912b", opacity=0.3)
DIEL_RO3206 = Material(er=6.15, tand=0.0027, color="#21912b", opacity=0.3)
DIEL_RO3210 = Material(er=10.2, tand=0.0027, color="#21912b", opacity=0.3)
DIEL_RO3730 = Material(er=3.00, tand=0.0016, color="#21912b", opacity=0.3)
DIEL_RO4003C = Material(er=3.38, tand=0.0029, color="#21912b", opacity=0.3)
DIEL_RO4350B = Material(er=3.48, tand=0.0037, color="#21912b", opacity=0.3)
DIEL_RO4350B_TX = Material(er=3.48, tand=0.0034, color="#21912b", opacity=0.3)
DIEL_RO4360 = Material(er=6.15, tand=0.0038, color="#21912b", opacity=0.3)
DIEL_RO4533 = Material(er=3.30, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_RO4534 = Material(er=3.40, tand=0.0027, color="#21912b", opacity=0.3)
DIEL_RO4535 = Material(er=3.50, tand=0.0037, color="#21912b", opacity=0.3)
DIEL_RO4730 = Material(er=3.00, tand=0.0033, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_5870 = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_5880 = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_5880LZ = Material(er=1.96, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6002 = Material(er=2.94, tand=0.0012, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6006 = Material(er=6.15, tand=0.0027, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6010_2LM = Material(er=10.2, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6035HTC = Material(er=3.50, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6202 = Material(er=2.94, tand=0.0015, color="#21912b", opacity=0.3)
DIEL_RT_Duroid_6202PR = Material(er=2.90, tand=0.002, color="#21912b", opacity=0.3)
DIEL_SYRON_70000_002IN_Thick = Material(er=3.4, tand=0.0045, color="#21912b", opacity=0.3)
DIEL_SYRON_71000_004IN_THICK = Material(er=3.39, tand=0.005, color="#21912b", opacity=0.3)
DIEL_SYRON_71000INCH = Material(er=3.61, tand=0.006, color="#21912b", opacity=0.3)
DIEL_TACLAMPLUS= Material(er=2.10, tand=0.0004, color="#21912b", opacity=0.3)
DIEL_TC350 = Material(er=3.50, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TC600 = Material(er=6.15, tand=0.002, color="#21912b", opacity=0.3)
DIEL_THETA = Material(er=3.85, tand=0.0123, color="#21912b", opacity=0.3)
DIEL_TLA_6 = Material(er=2.62, tand=0.0017, color="#21912b", opacity=0.3)
DIEL_TLC_27 = Material(er=2.75, tand=0.003, color="#21912b", opacity=0.3)
DIEL_TLC_30 = Material(er=3.00, tand=0.003, color="#21912b", opacity=0.3)
DIEL_TLC_32 = Material(er=3.20, tand=0.003, color="#21912b", opacity=0.3)
DIEL_TLC_338 = Material(er=3.38, tand=0.0034, color="#21912b", opacity=0.3)
DIEL_TLC_35 = Material(er=3.50, tand=0.0037, color="#21912b", opacity=0.3)
DIEL_TLE_95 = Material(er=2.95, tand=0.0028, color="#21912b", opacity=0.3)
DIEL_TLF_34 = Material(er=3.40, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TLF_35 = Material(er=3.50, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TLG_29 = Material(er=2.87, tand=0.0027, color="#21912b", opacity=0.3)
DIEL_TLG_30 = Material(er=3, tand=0.0038, color="#21912b", opacity=0.3)
DIEL_TLP_3 = Material(er=2.33, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLP_5 = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLP_5A = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLT_0 = Material(er=2.45, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLT_6 = Material(er=2.65, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLT_7 = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLT_8 = Material(er=2.55, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLT_9 = Material(er=2.50, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLX_0 = Material(er=2.45, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLX_6 = Material(er=2.65, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLX_7 = Material(er=2.60, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLX_8 = Material(er=2.55, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLX_9 = Material(er=2.50, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_TLY_3 = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3)
DIEL_TLY_3F = Material(er=2.33, tand=0.0012, color="#21912b", opacity=0.3)
DIEL_TLY_5 = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLY_5_L = Material(er=2.20, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLY_5A = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TLY_5AL = Material(er=2.17, tand=0.0009, color="#21912b", opacity=0.3)
DIEL_TMM_10 = Material(er=9.20, tand=0.0022, color="#21912b", opacity=0.3)
DIEL_TMM_10i = Material(er=9.80, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TMM_3 = Material(er=3.27, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TMM_4 = Material(er=4.50, tand=0.002, color="#21912b", opacity=0.3)
DIEL_TMM_6 = Material(er=6.00, tand=0.0023, color="#21912b", opacity=0.3)
DIEL_TRF_41 = Material(er=4.10, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_TRF_43 = Material(er=4.30, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_TRF_45 = Material(er=4.50, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_TSM_26 = Material(er=2.60, tand=0.0014, color="#21912b", opacity=0.3)
DIEL_TSM_29 = Material(er=2.94, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_TSM_30 = Material(er=3.00, tand=0.0013, color="#21912b", opacity=0.3)
DIEL_TSM_DS = Material(er=2.85, tand=0.001, color="#21912b", opacity=0.3)
DIEL_TSM_DS3 = Material(er=3.00, tand=0.0011, color="#21912b", opacity=0.3)
DIEL_ULTRALAM_2000 = Material(er=2.4, tand=0.0019, color="#21912b", opacity=0.3)
DIEL_ULTRALAM_3850 = Material(er=2.9, tand=0.0025, color="#21912b", opacity=0.3)
DIEL_XT_Duroid_80000_002IN_Thick = Material(er=3.23, tand=0.0035, color="#21912b", opacity=0.3)
DIEL_XT_Duroid_8100 = Material(er=3.54, tand=0.0049, color="#21912b", opacity=0.3)
DIEL_XT_Duroid_81000_004IN_Thick = Material(er=3.32, tand=0.0038, color="#21912b", opacity=0.3)

# Legacy FR Materials
DIEL_FR1 = Material(er=4.8,  tand=0.025, color="#3c9747", opacity=0.3) # Paper + phenolic resin
DIEL_FR2 = Material(er=4.8,  tand=0.02,  color="#3c9747", opacity=0.3) # Paper + phenolic resin
DIEL_FR3 = Material(er=4.5,  tand=0.02,  color="#2b7a4b", opacity=0.3) # Paper + epoxy resin
DIEL_FR4 = Material(er=4.4,  tand=0.015, color="#1e8449", opacity=0.3) # Woven glass + epoxy resin (industry standard)
DIEL_FR5 = Material(er=4.2,  tand=0.012, color="#156e38", opacity=0.3) # Woven glass + high-temp epoxy resin
DIEL_FR6 = Material(er=5.2,  tand=0.030, color="#145a32", opacity=0.3) # Paper + unknown resin, poor thermal performance

# Magnetic Materials
MU_METAL  = Material(cond=1.0e6, ur=200000, color="#666680", opacity=0.3)


############################################################
#                           FOAMS                          #
############################################################

FOAM_ROHACELL_31    = Material(er=1.05,  tand=0.0005,  color="#f0e1a1", opacity=0.15) # PMI-based structural foam
FOAM_ROHACELL_51    = Material(er=1.07,  tand=0.0006,  color="#f0dea0", opacity=0.15) # denser version
FOAM_ROHACELL_71    = Material(er=1.10,  tand=0.0007,  color="#e5d199", opacity=0.15)
FOAM_PEI       = Material(er=1.15,  tand=0.0035,  color="#e0b56f", opacity=0.15) # polyetherimide-based foam
FOAM_PMI       = Material(er=1.10,  tand=0.0008,  color="#d9c690", opacity=0.15) # polymethacrylimide
FOAM_PVC       = Material(er=1.20,  tand=0.0040,  color="#cccccc", opacity=0.15)
FOAM_EPS            = Material(er=1.03,  tand=0.0050,  color="#f7f7f7", opacity=0.15) # expanded polystyrene
FOAM_XPS            = Material(er=1.05,  tand=0.0030,  color="#e0e0e0", opacity=0.15) # extruded polystyrene
FOAM_PU        = Material(er=1.10,  tand=0.0080,  color="#d0d0d0", opacity=0.15) # polyurethane foam
FOAM_GLAS       = Material(er=3.10,  tand=0.0050,  color="#888888", opacity=0.15) # cellular glass, denser
FOAM_AIREX_C70      = Material(er=1.10,  tand=0.0010,  color="#f7e7a3", opacity=0.15) # PET closed cell
FOAM_AIREX_T92      = Material(er=1.10,  tand=0.0020,  color="#f6d08a", opacity=0.15) # higher strength PET
FOAM_PVC_CORECELL   = Material(er=1.56,  tand=0.0025,  color="#aaaaaa", opacity=0.15) # structural core PVC




import numpy as np
import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from typing import (
    Union, Sequence, Callable, List, Optional, Tuple
)
from cycler import cycler

#_colors = plt.rcParams['axes.prop_cycle'].by_key()['color']

EMERGE_COLORS = ["#1A14CE", "#D54A09", "#1F82A6", "#D3107B", "#119D40"]
EMERGE_CYCLER = cycler(color=EMERGE_COLORS)
plt.rc('axes', prop_cycle=EMERGE_CYCLER)

ggplot_styles = {
    "axes.edgecolor": "000000",
    "axes.facecolor": "F2F2F2",
    "axes.grid": True,
    "axes.grid.which": "both",
    "axes.spines.left": True,
    "axes.spines.right": True,
    "axes.spines.top": True,
    "axes.spines.bottom": True,
    "grid.color": "A0A0A0",
    "grid.linewidth": "0.8",
    "xtick.color": "555555",
    "xtick.major.bottom": True,
    "xtick.minor.bottom": False,
    "ytick.color": "555555",
    "ytick.major.left": True,
    "ytick.minor.left": False,
    "lines.linewidth": 2,
}

plt.rcParams.update(ggplot_styles)

def _gen_grid(xs: tuple, ys: tuple, N = 201) -> list[tuple[np.ndarray, np.ndarray]]:
    """Generate a grid of lines for the Smith Chart

    Args:
        xs (tuple): Tuple containing the x-axis values
        ys (tuple): Tuple containing the y-axis values
        N (int, optional): Number Used. Defaults to 201.

    Returns:
        list[np.ndarray]: List of lines
    """    
    xgrid = np.arange(xs[0], xs[1]+xs[2], xs[2])
    ygrid = np.arange(ys[0], ys[1]+ys[2], ys[2])
    xsmooth = np.logspace(np.log10(xs[0]+1e-8), np.log10(xs[1]), N)
    ysmooth = np.logspace(np.log10(ys[0]+1e-8), np.log10(ys[1]), N)
    ones = np.ones((N,))
    lines = []
    for x in xgrid:
        lines.append((x*ones, ysmooth))
        lines.append((x*ones, -ysmooth))
    for y in ygrid:
        lines.append((xsmooth, y*ones))
        lines.append((xsmooth, -y*ones))
        
    return lines

def _generate_grids(orders = (0, 0.5, 1, 2, 5, 10, 50,1e5), N=201) -> list[tuple[np.ndarray, np.ndarray]]:
    """Generate the grid for the Smith Chart

    Args:
        orders (tuple, optional): Locations for Smithchart Lines. Defaults to (0, 0.5, 1, 2, 5, 10, 50,1e5).
        N (int, optional): N distrectization points. Defaults to 201.

    Returns:
        list[tuple[np.ndarray, np.ndarray]]: List of axes lines
    """    
    lines = []
    xgrids = orders
    for o1, o2 in zip(xgrids[:-1], xgrids[1:]):
        step = o2/10
        lines += _gen_grid((0, o2, step), (0, o2, step), N)   
    return lines

def _smith_transform(lines: list[tuple[np.ndarray, np.ndarray]]) -> list[tuple[np.ndarray, np.ndarray]]:
    """Executes the Smith Transform on a list of lines

    Args:
        lines (list[tuple[np.ndarray, np.ndarray]]): List of lines

    Returns:
        list[tuple[np.ndarray, np.ndarray]]: List of transformed lines
    """    
    new_lines = []
    for line in lines:
        x, y = line
        z = x + 1j*y
        new_z = (z-1)/(z+1)
        new_x = new_z.real
        new_y = new_z.imag
        new_lines.append((new_x, new_y))
    return new_lines

def hintersections(x: np.ndarray, y: np.ndarray, level: float) -> list[float]:
    """Find the intersections of a line with a level

    Args:
        x (np.ndarray): X-axis values
        y (np.ndarray): Y-axis values
        level (float): Level to intersect

    Returns:
        list[float]: List of x-values where the intersection occurs
    """      
    y1 = y[:-1] - level
    y2 = y[1:] - level
    ycross = y1 * y2
    id1 = np.where(ycross < 0)[0]
    id2 = id1 + 1
    x1 = x[id1]
    x2 = x[id2]
    y1 = y[id1] - level
    y2 = y[id2] - level
    a = (y2 - y1) / (x2 - x1)
    b = y1 - a * x1
    xcross = list(-b / a)
    xlevel = list(x[np.where(y == level)])
    return xcross + xlevel



def plot(
    x: np.ndarray,
    y: Union[np.ndarray, Sequence[np.ndarray]],
    grid: bool = True,
    labels: Optional[List[str]] = None,
    xlabel: str = "x",
    ylabel: str = "y",
    linestyles: Union[str, List[str]] = "-",
    linewidth: float = 2.0,
    markers: Optional[Union[str, List[Optional[str]]]] = None,
    logx: bool = False,
    logy: bool = False,
    transformation: Optional[Callable[[np.ndarray], np.ndarray]] = None,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None
) -> None:
    """
    Plot one or more y‐series against a common x‐axis, with extensive formatting options.

    Parameters
    ----------
    x : np.ndarray
        1D array of x‐values.
    y : np.ndarray or sequence of np.ndarray
        Either a single 1D array of y‐values, or a sequence of such arrays.
    grid : bool, default True
        Whether to show the grid.
    labels : list of str, optional
        One label per series. If None, no legend is drawn.
    xlabel : str, default "x"
        Label for the x‐axis.
    ylabel : str, default "y"
        Label for the y‐axis.
    linestyles : str or list of str, default "-"
        Matplotlib linestyle(s) for each series.
    linewidth : float, default 2.0
        Line width for all series.
    markers : str or list of str or None, default None
        Marker style(s) for each series. If None, no markers.
    logx : bool, default False
        If True, set x‐axis to logarithmic scale.
    logy : bool, default False
        If True, set y‐axis to logarithmic scale.
    transformation : callable, optional
        Function `f(y)` to transform each y‐array before plotting.
    xlim : tuple (xmin, xmax), optional
        Limits for the x‐axis.
    ylim : tuple (ymin, ymax), optional
        Limits for the y‐axis.
    title : str, optional
        Figure title.
    """
    # Ensure y_list is a list of arrays
    if isinstance(y, np.ndarray):
        y_list = [y]
    else:
        y_list = list(y)

    n_series = len(y_list)

    # Prepare labels, linestyles, markers
    if labels is not None and len(labels) != n_series:
        raise ValueError("`labels` length must match number of y‐series")
    # Turn single styles into lists of length n_series
    def _broadcast(param, default):
        if isinstance(param, list):
            if len(param) != n_series:
                raise ValueError(f"List length of `{param}` must match number of series")
            return param
        else:
            return [param] * n_series

    linestyles = _broadcast(linestyles, "-")
    markers = _broadcast(markers, None) if markers is not None else [None] * n_series

    # Apply transformation if given
    if transformation is not None:
        y_list = [trans(y_i) for trans, y_i in zip([transformation]*n_series, y_list)]

    # Create plot
    fig, ax = plt.subplots()
    for i, y_i in enumerate(y_list):
        ax.plot(
            x, y_i,
            linestyle=linestyles[i],
            linewidth=linewidth,
            marker=markers[i],
            label=(labels[i] if labels is not None else None)
        )

    # Axes scales
    if logx:
        ax.set_xscale("log")
    if logy:
        ax.set_yscale("log")

    # Grid, labels, title
    ax.grid(grid)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title is not None:
        ax.set_title(title)

    # Limits
    if xlim is not None:
        ax.set_xlim(*xlim)
    if ylim is not None:
        ax.set_ylim(*ylim)

    # Legend
    if labels is not None:
        ax.legend()

    plt.show()

def smith(f: np.ndarray, S: np.ndarray) -> None:
    """ Plot the Smith Chart

    Args:
        f (np.ndarray): Frequency vector (Not used yet)
        S (np.ndarray): S-parameters to plot
    """     
    if not isinstance(S, list):
        Ss = [S]
    else:
        Ss = S
    
    fig, ax = plt.subplots()
    for line in _smith_transform(_generate_grids()):
        ax.plot(line[0], line[1], color='grey', alpha=0.3, linewidth=0.7)
    p = np.linspace(0,2*np.pi,101)
    ax.plot(np.cos(p), np.sin(p), color='black', alpha=0.5)
    # Add important numbers for the Impedance Axes
    # X and Y values (0, 0.5, 1, 2, 10, 50)
    for i in [0, 0.2, 0.5, 1, 2, 10]:
        z = i + 1j*0
        G = (z-1)/(z+1)
        ax.annotate(f"{i}", (G.real, G.imag), color='black', fontsize=8)
    for i in [0, 0.2, 0.5, 1, 2, 10]:
        z = 0 + 1j*i
        G = (z-1)/(z+1)
        ax.annotate(f"{i}", (G.real, G.imag), color='black', fontsize=8)       
        ax.annotate(f"{-i}", (G.real, -G.imag), color='black', fontsize=8)  
    for s in Ss:
        ax.plot(s.real, s.imag, color='blue')
    ax.grid(False)
    ax.axis('equal')
    plt.show()

def plot_sp(f: np.ndarray, S: list[np.ndarray] | np.ndarray, 
                      dblim=[-40, 5], 
                    xunit="GHz", 
                    levelindicator: int | float | None = None, 
                    noise_floor=-150, 
                    fill_areas: list[tuple] | None = None, 
                    spec_area: list[tuple[float,...]] | None = None,
                    unwrap_phase=False, 
                    logx: bool = False,
                    labels: list[str] | None = None,
                    linestyles: list[str] | None = None,
                    colorcycle: list[int] | None = None,
                    filename: str | None = None,
                    show_plot: bool = True,
                    figdata: tuple | None = None) -> tuple[plt.Figure, plt.Axes, plt.Axes]:
    """Plot S-parameters in dB and phase

    Args:
        f (np.ndarray): Frequency vector
        S (list[np.ndarray] | np.ndarray): S-parameters to plot (list or single array)
        dblim (list, optional): Decibel y-axis limit. Defaults to [-80, 5].
        xunit (str, optional): Frequency unit. Defaults to "GHz".
        levelindicator (int | float, optional): Level at which annotation arrows will be added. Defaults to None.
        noise_floor (int, optional): Artificial random noise floor level. Defaults to -150.
        fill_areas (list[tuple], optional): Regions to fill (fmin, fmax). Defaults to None.
        spec_area (list[tuple[float]], optional): _description_. Defaults to None.
        unwrap_phase (bool, optional): If or not to unwrap the phase data. Defaults to False.
        logx (bool, optional): Whether to use logarithmic frequency axes. Defaults to False.
        labels (list[str], optional): A lists of labels to use. Defaults to None.
        linestyles (list[str], optional): The linestyle to use (list or single string). Defaults to None.
        colorcycle (list[int], optional): A list of colors to use. Defaults to None.
        filename (str, optional): The filename (will automatically save). Defaults to None.
        show_plot (bool, optional): If or not to show the resulting plot. Defaults to True.
    """    
    if not isinstance(S, list):
        Ss = [S]
    else:
        Ss = S

    if linestyles is None:
        linestyles = ['-' for _ in S]

    if colorcycle is None:
        colorcycle = [i for i, S in enumerate(S)]

    unitdivider: dict[str, float] = {"MHz": 1e6, "GHz": 1e9, "kHz": 1e3}
    fnew = f / unitdivider[xunit]

    if figdata is None:
        # Create two subplots: one for magnitude and one for phase
        fig, (ax_mag, ax_phase) = plt.subplots(2, 1, sharex=False, gridspec_kw={'height_ratios': [3, 1]})
        fig.subplots_adjust(hspace=0.3)
    else:
        fig, ax_mag, ax_phase = figdata
    minphase, maxphase = -180, 180

    maxy = 0
    for s, ls, cid in zip(Ss, linestyles, colorcycle):
        # Calculate and plot magnitude in dB
        SdB = 20 * np.log10(np.abs(s) + 10**(noise_floor/20) * np.random.rand(*s.shape) + 10**((noise_floor-30)/20))
        ax_mag.plot(fnew, SdB, label="Magnitude (dB)", linestyle=ls, color=EMERGE_COLORS[cid % len(EMERGE_COLORS)])
        if np.max(SdB) > maxy:
            maxy = np.max(SdB)
        # Calculate and plot phase in degrees
        phase = np.angle(s, deg=True)
        if unwrap_phase:
            phase = np.unwrap(phase, period=360)
            minphase = min(np.min(phase), minphase)
            maxphase = max(np.max(phase), maxphase)
        ax_phase.plot(fnew, phase, label="Phase (degrees)", linestyle=ls, color=EMERGE_COLORS[cid % len(EMERGE_COLORS)])

        # Annotate level indicators if specified
        if isinstance(levelindicator, (int, float)) and levelindicator is not None:
            lvl = levelindicator
            fcross = hintersections(fnew, SdB, lvl)
            for fs in fcross:
                ax_mag.annotate(
                    f"{str(fs)[:4]}{xunit}",
                    xy=(fs, lvl),
                    xytext=(fs + 0.08 * (max(f) - min(f)) / unitdivider[xunit], lvl),
                    arrowprops=dict(facecolor="black", width=1, headwidth=5),
                )
    if fill_areas is not None:
        for fmin, fmax in fill_areas:
            f1 = fmin / unitdivider[xunit]
            f2 = fmax / unitdivider[xunit]
            ax_mag.fill_between([f1, f2], dblim[0], dblim[1], color='grey', alpha= 0.2)
            ax_phase.fill_between([f1, f2], minphase, maxphase, color='grey', alpha= 0.2)
    if spec_area is not None:
        for fmin, fmax, vmin, vmax in spec_area:
            f1 = fmin / unitdivider[xunit]
            f2 = fmax / unitdivider[xunit]
            ax_mag.fill_between([f1, f2], vmin,vmax, color='red', alpha=0.2)
    # Configure magnitude plot (ax_mag)
    ax_mag.set_ylabel("Magnitude (dB)")
    ax_mag.set_xlabel(f"Frequency ({xunit})")
    ax_mag.axis([min(fnew), max(fnew), dblim[0], max(maxy*1.1,dblim[1])]) # type: ignore
    ax_mag.axhline(y=0, color="k", linewidth=1)
    ax_mag.xaxis.set_minor_locator(tck.AutoMinorLocator(2))
    ax_mag.yaxis.set_minor_locator(tck.AutoMinorLocator(2))
    # Configure phase plot (ax_phase)
    ax_phase.set_ylabel("Phase (degrees)")
    ax_phase.set_xlabel(f"Frequency ({xunit})")
    ax_phase.axis([min(fnew), max(fnew), minphase, maxphase]) # type: ignore
    ax_phase.xaxis.set_minor_locator(tck.AutoMinorLocator(2))
    ax_phase.yaxis.set_minor_locator(tck.AutoMinorLocator(2))
    if logx:
        ax_mag.set_xscale('log')
        ax_phase.set_xscale('log')
    if labels is not None:
        ax_mag.legend(labels)
        ax_phase.legend(labels)
    if show_plot:
        plt.show()
    if filename is not None:
        fig.savefig(filename)

    return fig, ax_mag, ax_phase

    
def plot_ff(
    theta: np.ndarray,
    E: Union[np.ndarray, Sequence[np.ndarray]],
    grid: bool = True,
    dB: bool = False,
    labels: Optional[List[str]] = None,
    xlabel: str = "Theta (rad)",
    ylabel: str = "|E|",
    linestyles: Union[str, List[str]] = "-",
    linewidth: float = 2.0,
    markers: Optional[Union[str, List[Optional[str]]]] = None,
    xlim: Optional[Tuple[float, float]] = None,
    ylim: Optional[Tuple[float, float]] = None,
    title: Optional[str] = None
) -> None:
    """
    Far-field rectangular plot of E-field magnitude vs angle.

    Parameters
    ----------
    theta : np.ndarray
        Angle array (radians).
    E : np.ndarray or sequence of np.ndarray
        Complex E-field samples; magnitude will be plotted.
    grid : bool
        Show grid.
    labels : list of str, optional
        Series labels.
    xlabel, ylabel : str
        Axis labels.
    linestyles, linewidth, markers : styling parameters.
    xlim, ylim : tuple, optional
        Axis limits.
    title : str, optional
        Plot title.
    """
    # Prepare data series
    if isinstance(E, np.ndarray):
        E_list = [E]
    else:
        E_list = list(E)
    n_series = len(E_list)

    # Style broadcasting
    def _broadcast(param, default):
        if isinstance(param, list):
            if len(param) != n_series:
                raise ValueError(f"List length of `{param}` must match number of series")
            return param
        else:
            return [param] * n_series

    linestyles = _broadcast(linestyles, "-")
    markers = _broadcast(markers, None) if markers is not None else [None] * n_series

    fig, ax = plt.subplots()
    for i, Ei in enumerate(E_list):
        mag = np.abs(Ei)
        if dB:
            mag = 20*np.log10(mag)
        ax.plot(
            theta, mag,
            linestyle=linestyles[i],
            linewidth=linewidth,
            marker=markers[i],
            label=(labels[i] if labels else None)
        )

    ax.grid(grid)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    if title:
        ax.set_title(title)
    if xlim:
        ax.set_xlim(*xlim)
    if ylim:
        ax.set_ylim(*ylim)
    if labels:
        ax.legend()

    plt.show()


def plot_ff_polar(
    theta: np.ndarray,
    E: Union[np.ndarray, Sequence[np.ndarray]],
    labels: Optional[List[str]] = None,
    linestyles: Union[str, List[str]] = "-",
    linewidth: float = 2.0,
    markers: Optional[Union[str, List[Optional[str]]]] = None,
    zero_location: str = 'N',
    clockwise: bool = False,
    rlabel_angle: float = 45,
    title: Optional[str] = None
) -> None:
    """
    Far-field polar plot of E-field magnitude vs angle.

    Parameters
    ----------
    theta : np.ndarray
        Angle array (radians).
    E : np.ndarray or sequence of np.ndarray
        Complex E-field samples; magnitude will be plotted.
    labels : list of str, optional
        Series labels.
    linestyles, linewidth, markers : styling parameters.
    zero_location : str
        Theta zero location (e.g. 'N', 'E').
    clockwise : bool
        If True, theta increases clockwise.
    rlabel_angle : float
        Position (deg) of radial labels.
    title : str, optional
        Plot title.
    """
    # Prepare data series
    if isinstance(E, np.ndarray):
        E_list = [E]
    else:
        E_list = list(E)
    n_series = len(E_list)

    # Style broadcasting
    def _broadcast(param, default):
        if isinstance(param, list):
            if len(param) != n_series:
                raise ValueError(f"List length of `{param}` must match number of series")
            return param
        else:
            return [param] * n_series

    linestyles = _broadcast(linestyles, "-")
    markers = _broadcast(markers, None) if markers is not None else [None] * n_series

    fig, ax = plt.subplots(subplot_kw={'projection': 'polar'})
    ax.set_theta_zero_location(zero_location) # type: ignore
    ax.set_theta_direction(-1 if clockwise else 1) # type: ignore
    ax.set_rlabel_position(rlabel_angle) # type: ignore

    for i, Ei in enumerate(E_list):
        mag = np.abs(Ei)
        ax.plot(
            theta, mag,
            linestyle=linestyles[i],
            linewidth=linewidth,
            marker=markers[i],
            label=(labels[i] if labels else None)
        )

    if title:
        ax.set_title(title, va='bottom')
    if labels:
        ax.legend(loc='upper right', bbox_to_anchor=(1.1, 1.1))

    plt.show()



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.
from __future__ import annotations
import os
import sys
import ctypes
import re
import site
from ctypes.util import find_library
from enum import Enum
import numpy as np
from scipy import sparse # type: ignore
from pathlib import Path
from typing import Iterable, Iterator
import pickle
from loguru import logger

############################################################
#                        ERROR CODES                       #
############################################################

PARDISO_ERROR_CODES = {
     0: "No error.",
    -1: "Input inconsistent.",
    -2: "Not enough memory.",
    -3: "Reordering problem.",
    -4: "Zero pivot, numerical fac. or iterative refinement problem.",
    -5: "Unclassified (internal) error.",
    -6: "Preordering failed (matrix types 11(real and nonsymmetric), 13(complex and nonsymmetric) only).",
    -7: "Diagonal Matrix problem.",
    -8: "32-bit integer overflow problem.",
   -10: "No license file pardiso.lic found.",
   -11: "License is expired.",
   -12: "Wrong username or hostname.",
  -100: "Reached maximum number of Krylov-subspace iteration in iterative solver.",
  -101: "No sufficient convergence in Krylov-subspace iteration within 25 iterations.",
  -102: "Error in Krylov-subspace iteration.",
  -103: "Bread-Down in Krylov-subspace iteration",
}


############################################################
#               FINDING THE PARDISO DLL FILES              #
############################################################


#: Environment variable that overrides automatic searching

ENV_VAR = "EMERGE_PARDISO_PATH"


def _candidate_dirs() -> Iterable[Path]:
    """Return directories in which to look for MKL."""
    # Ordered from most to least likely
    seen: set[Path] = set()

    for p in (                         # likely “local” env first
        Path(sys.prefix),
        Path(getattr(sys, "base_prefix", sys.prefix)),
        Path(str(site.USER_BASE)),
        *(Path(x) for x in os.getenv("LD_LIBRARY_PATH", "").split(":") if x),
    ):
        if p not in seen:
            seen.add(p)
            yield p

def _search_mkl() -> Iterator[Path]:
    """Yield candidate MKL library paths, shortest first."""
    pattern = {
        "win32": r"^mkl_rt.*\.dll$",
        "darwin": r"^libmkl_rt(\.\d+)*\.dylib$",
        "linux": r"^libmkl_rt(\.so(\.\d+)*)?$",
    }.get(sys.platform, r"^libmkl_rt")

    regex = re.compile(pattern, re.IGNORECASE)

    for base in _candidate_dirs():
        for path in sorted(base.rglob("**/*mkl_rt*"), key=lambda p: len(str(p))):
            if regex.match(path.name):
                yield path

def cache_path_result(tag: str, compute_fn, force: bool = False) -> str:
    """
    Retrieve a cached Path object or compute it and store it.
    
    Parameters
    ----------
    tag : str
        Cache key.
    compute_fn : callable
        Callable that returns a Path.
    force : bool
        If True, bypass and overwrite the cache.
    """
    cache_dir = Path(__file__).parent / "__pycache__"
    cache_dir.mkdir(exist_ok=True)
    cache_file = cache_dir / f"{tag}.pkl"

    if not force and cache_file.exists():
        with open(cache_file, "rb") as f:
            filename = pickle.load(f)
            logger.debug(f"Using cached MKL file: {filename}")
            return str(filename)

    result = compute_fn()
    with open(cache_file, "wb") as f:
        pickle.dump(result, f)
    return str(result)

def search_mkl() -> str:
    """Searches for the file path of the PARDISO MKL executable

    Returns:
        str: The filepath
    """
    logger.debug('Searching for MKL executable...')
    for candidate in _search_mkl():
        try:
            ctypes.CDLL(str(candidate))
        except OSError:
            continue  # try the next one
    logger.debug(f'Executable found: {candidate}')
    return str(candidate)

def load_mkl() -> ctypes.CDLL:
    """Locate and load **mkl_rt**; raise ImportError on failure."""
    # 1 explicit override
    override = os.getenv(ENV_VAR)
    if override:
        try:
            return ctypes.CDLL(override)
        except OSError as e:
            raise ImportError(f"{override!r} could not be loaded: {e}") from None

    # 2 system utility (cheap)
    lib = find_library("mkl_rt")
    if lib:
        try:
            return ctypes.CDLL(lib)
        except OSError:
            pass
    
    # 2 system utility (cheap)
    lib = find_library("mkl_rt.1")
    if lib:
        try:
            return ctypes.CDLL(lib)
        except OSError:
            pass

    # 3 filesystem walk (expensive, but last resort)
    try:
        filename = cache_path_result('mkl_file',search_mkl)
        return ctypes.CDLL(filename)
    except OSError:
        logger.warning('File name {filename} is no longer valid. Re-executing MKL search.')
        filename = cache_path_result('mkl_file',search_mkl,force=True)
        return ctypes.CDLL(filename)

    raise ImportError(
        "Shared library *mkl_rt* not found. "
        f"Set the environment variable {ENV_VAR} to its full path if it is in a non-standard location."
    )

    

############################################################
#                  ALL C-TYPE DEFINITIONS                 #
############################################################


class MKL_Complex16(ctypes.Structure):
    _fields_ = [("real", ctypes.c_double),
                ("imag", ctypes.c_double)]

CINT64 = ctypes.c_int64
CINT32 = ctypes.c_int32
CFLOAT32 = ctypes.c_float
CFLOAT64 = ctypes.c_double

CPX16_P = ctypes.POINTER(MKL_Complex16)
CINT64_P = ctypes.POINTER(CINT64)
CINT32_P = ctypes.POINTER(CINT32)
CNONE_P = ctypes.POINTER(None)
CFLOAT32_P = ctypes.POINTER(CFLOAT32)
CFLOAT64_P = ctypes.POINTER(CFLOAT64)
VOID_SIZE = ctypes.sizeof(ctypes.c_void_p)

PT_A: type = CINT32
PT_B: type = np.int32

if VOID_SIZE == 8:
    PT_A = CINT64
    PT_B = np.int64
elif VOID_SIZE == 4:
    PT_A = CINT32
    PT_B = np.int32

def c_int(value: int):
    return ctypes.byref(ctypes.c_int32(value))

PARDISO_ARG_TYPES: tuple = (ctypes.POINTER(PT_A),CINT32_P,CINT32_P,
    CINT32_P,CINT32_P,CINT32_P,CNONE_P,CINT32_P,CINT32_P,
    CINT32_P,CINT32_P,CINT32_P,CINT32_P,CNONE_P,CNONE_P,CINT32_P,
)


############################################################
#                  PARDISO CONFIGURATIONS                 #
############################################################


class PARDISOMType(Enum):
    REAL_SYM_STRUCT = 1
    REAL_SYM_POSDEF = 2
    REAL_SYM_INDEF = -2
    COMP_SYM_STRUCT = 3
    COMP_HERM_POSDEF = 4
    COMP_HERM_INDEF = -4
    COMP_SYM = 6
    REAL_NONSYM = 11
    COMP_NONSYM = 13

class PARDISOPhase(Enum):
    SYMBOLIC_FACTOR = 11
    NUMERIC_FACTOR = 12
    NUMERIC_SOLVE = 33
    


############################################################
#                   GENERIC MATRIX CLASS                  #
############################################################


class SolveMatrix:
    def __init__(self, A: sparse.csr_matrix):
        A = A.tocsr()
        if not A.has_sorted_indices:
            A.sort_indices()

        if (A.getnnz(axis=1) == 0).any() or (A.getnnz(axis=0) == 0).any():
            raise ValueError('Matrix A is singular, because it contains empty rows or columns')

        if A.dtype in (np.float16, np.float32, np.float64):
            A = A.astype(np.float64)
        elif np.iscomplexobj(A):
            A = A.astype(np.complex128)
        self.mat: sparse.csr_matrix = A

        self.factorized_symb: bool = False
        self.factorized_num: bool = False

    
    def is_same(self, matrix: SolveMatrix):
        return id(self.mat) is id(matrix.mat)

    def same_structure(self, matrix: SolveMatrix):
        return np.all(matrix.mat.indices == self.mat.indices) & np.all(matrix.mat.indptr == self.mat.indptr)

    @property
    def zerovec(self):
        return np.zeros_like((self.mat.shape[0], 1), dtype=self.mat.dtype)


############################################################
#                   THE PARDISO INTERFACE                  #
############################################################

class PardisoInterface:

    def __init__(self):

        self.libmkl: ctypes.CDLL = load_mkl()
        self._pardiso_interface = self.libmkl.pardiso

        self._pardiso_interface.restype = None
        self._pardiso_interface.argtypes = PARDISO_ARG_TYPES

        self.PT = np.zeros(64, dtype=PT_B)
        self.IPARM = np.zeros(64, dtype=np.int32)
        self.PERM = np.zeros(0, dtype=np.int32)
        self.MATRIX_TYPE: PARDISOMType = None
        self.complex: bool = False
        self.mat_structure: np.ndarray = None
        self.message_level: int = 0

        self.matrix: SolveMatrix = None
        self.factored_matrix: SolveMatrix = None

        self.configure_solver()

    def _configure(self, A: sparse.csr_matrix) -> None:
        """Configures the solver for the appropriate data type (float/complex)

        Args:
            A (sparse.csr_matrix): The sparse matrix to solve
        """
        if np.iscomplexobj(A):
            self.MATRIX_TYPE = PARDISOMType.COMP_SYM_STRUCT
            self.complex = True
        else:
            self.MATRIX_TYPE = PARDISOMType.REAL_SYM_STRUCT
            self.complex = False

    def _prepare_B(self, b: np.ndarray | sparse.sparray) -> np.ndarray:
        """Fixes the forcing-vector for the solution process

        Args:
            b (np.ndarray): The forcing vector in Ax=b

        Returns:
            np.ndarray: The prepared forcing-vector
        """
        if sparse.issparse(b):
            b = b.todense() # type: ignore
        
        if np.iscomplexobj(b):
            b = b.astype(np.complex128)
        else:
            b = b.astype(np.float64)
        return b
    
    def symbolic(self, A: sparse.csr_matrix) -> int:
        """Calls the Symbollic solve routinge

        Returns:
            int: The error code
        """
        
        self._configure(A)
        zerovec = np.zeros_like((A.shape[0], 1), dtype=A.dtype)
        _, error = self._call_solver(A, zerovec, phase=PARDISOPhase.SYMBOLIC_FACTOR)
        return error

    def numeric(self, A: sparse.csr_matrix) -> int:
        """Calls the Numeric solve routine

        Returns:
            int: The error code
        """

        self._configure(A)
        zerovec = np.zeros_like((A.shape[0], 1), dtype=A.dtype)
        _, error = self._call_solver(A, zerovec, phase=PARDISOPhase.NUMERIC_FACTOR)
        return error
        
    def solve(self, A: sparse.csr_matrix, b: np.ndarray) -> tuple[np.ndarray, int]:
        """ Solves the linear problem Ax=b with PARDISO

        Args:
            A (sparse.csr_matrix): The A-matrix
            b (np.ndarray): The b-vector

        Returns:
            tuple[np.ndarray, int]: The solution vector x, and error code
        """
        self._configure(A)
        b = self._prepare_B(b)
        x, error = self._call_solver(A, b, phase=PARDISOPhase.NUMERIC_SOLVE)
        return x, error

    def configure_solver(self, 
        perm_algo: int = 3,
        nthreads: int | None = None,
        user_perm: int = 0,
        n_refine_steps: int = 0,
        pivot_pert: int = 13,
        weighted_matching: int = 2):
        """Configures the solver

        Args:
            perm_algo (int, optional): The permutation algorithm. Defaults to 3.
            nthreads (int, optional): The number of threads (Must be greater than OMP_NUM_THREADS). Defaults to None.
            user_perm (int, optional): 1, if a user permuation is provided (not supported yet). Defaults to 0.
            n_refine_steps (int, optional): Number of refinement steps. Defaults to 0.
            pivot_pert (int, optional): _description_. Defaults to 13.
            weighted_matching (int, optional): weighted matching mode. Defaults to 2.
        """
        if nthreads is None:
            nthreads = int(os.environ.get('OMP_NUM_THREADS', default="4"))

        self.IPARM[1] = perm_algo
        self.IPARM[2] = nthreads
        self.IPARM[4] = user_perm
        self.IPARM[7] = n_refine_steps
        self.IPARM[9] = pivot_pert
        self.IPARM[12] = weighted_matching

    def _call_solver(self, A: sparse.csr_matrix, b: np.ndarray, phase: PARDISOPhase) -> tuple[np.ndarray, int]:
        """Calls the PARDISO solver on linear problem Ax=b

        Args:
            A (sparse.csr_matrix): The A-matrix
            b (np.ndarray): The b-vector
            phase (PARDISOPhase): The solution phase

        Returns:
            tuple[np.ndarray, int]: The solution vector x and error code.
        """

        # Declare the empty vector
        x = np.zeros_like(b)
        error = ctypes.c_int32(0)
        
        # Up the pointers as PARDISO uses [1,...] indexing
        A_index_pointers = A.indptr + 1
        A_indices = A.indices + 1

        # Define the appropriate data type (complex vs real)

        if self.complex:
            VALUE_P = A.data.ctypes.data_as(CPX16_P)
            RHS_P = b.ctypes.data_as(CPX16_P)
            X_P = x.ctypes.data_as(CPX16_P)
        else:
            VALUE_P = A.data.ctypes.data_as(CFLOAT64_P)
            RHS_P = b.ctypes.data_as(CFLOAT64_P) # type: ignore
            X_P = x.ctypes.data_as(CFLOAT64_P) # type: ignore

        # Calls the pardiso function
        self._pardiso_interface(
            self.PT.ctypes.data_as(ctypes.POINTER(PT_A)),
            c_int(1),
            c_int(1),
            c_int(self.MATRIX_TYPE.value),
            c_int(phase.value),
            c_int(A.shape[0]),
            VALUE_P,
            A_index_pointers.ctypes.data_as(CINT32_P),
            A_indices.ctypes.data_as(CINT32_P),
            self.PERM.ctypes.data_as(CINT32_P),
            c_int(1),
            self.IPARM.ctypes.data_as(CINT32_P),
            c_int(self.message_level),
            RHS_P,
            X_P,
            ctypes.byref(error))
        
        # Returns the solution vector plus error code
        return np.ascontiguousarray(x), error.value

    def get_error(self, error: int) -> str:
        """Returns the PARDISO error description string given an error number

        Args:
            error (int): The error number

        Returns:
            str: A description string
        """
        return PARDISO_ERROR_CODES[error]

    def clear_memory(self):
        """Clear the memory of this solver plus the PARDISO process.
        """
        self.factorized_A = None
        self.matrix = None
        self.symbolic(sparse.csr_matrix((0,0), dtype=np.complex128), np.zeros(0, dtype=np.complex128))



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from numba_progress import ProgressBar
from ...mesh3d import SurfaceMesh
import numpy as np
from loguru import logger
from numba import c16, njit, prange, f8
from numba.types import Tuple as TupleType  # ty: ignore
from numba_progress.progress import ProgressBarType
from ...const import Z0

LR = 0.001

@njit(
    TupleType((c16[:, :], c16[:, :]))(
        c16[:, :],
        c16[:, :],
        f8[:, :],
        f8[:, :],
        f8[:, :],
        f8,
        ProgressBarType,
    ),
    parallel=True,
    fastmath=True,
    cache=True,
    nogil=True,
)
def stratton_chu_ff(Ein, Hin, vis, wns, tpout, k0, pgb):
    
    Ex = Ein[0, :].flatten()
    Ey = Ein[1, :].flatten()
    Ez = Ein[2, :].flatten()
    Hx = Hin[0, :].flatten()
    Hy = Hin[1, :].flatten()
    Hz = Hin[2, :].flatten()
    vx = vis[0, :].flatten()
    vy = vis[1, :].flatten()
    vz = vis[2, :].flatten()
    nx = wns[0, :].flatten()
    ny = wns[1, :].flatten()
    nz = wns[2, :].flatten()
    
    Emag = np.sqrt(np.abs(Ex)**2 + np.abs(Ey)**2 + np.abs(Ez)**2)
    
    Elevel = np.max(Emag) * LR
    ids = np.argwhere(Emag > Elevel)
    Nids = ids.shape[0]
    #iadd = NT // Nids
    Ex = Ex[Emag > Elevel]
    Ey = Ey[Emag > Elevel]
    Ez = Ez[Emag > Elevel]
    Hx = Hx[Emag > Elevel]
    Hy = Hy[Emag > Elevel]
    Hz = Hz[Emag > Elevel]
    vx = vx[Emag > Elevel]
    vy = vy[Emag > Elevel]
    vz = vz[Emag > Elevel]
    nx = nx[Emag > Elevel]
    ny = ny[Emag > Elevel]
    nz = nz[Emag > Elevel]

    thout = tpout[0, :]
    phout = tpout[1, :]

    rx = np.sin(thout) * np.cos(phout)
    ry = np.sin(thout) * np.sin(phout)
    rz = np.cos(thout)

    kx = k0 * rx
    ky = k0 * ry
    kz = k0 * rz

    N = tpout.shape[1]

    Eout = np.zeros((3, N)).astype(np.complex128)
    Hout = np.zeros((3, N)).astype(np.complex128)

    Eoutx = np.zeros((N,)).astype(np.complex128)
    Eouty = np.zeros((N,)).astype(np.complex128)
    Eoutz = np.zeros((N,)).astype(np.complex128)

    Q = np.complex128(-1j * k0 / (4 * np.pi))
    ii = np.complex128(1j)

    NxHx = ny * Hz - nz * Hy
    NxHy = nz * Hx - nx * Hz
    NxHz = nx * Hy - ny * Hx

    NxEx = ny * Ez - nz * Ey
    NxEy = nz * Ex - nx * Ez
    NxEz = nx * Ey - ny * Ex

    for j in prange(Nids): # ty: ignore
        xi = vx[j]
        yi = vy[j]
        zi = vz[j]
        G = np.exp(ii * (kx * xi + ky * yi + kz * zi))

        RxNxHx = ry * NxHz[j] - rz * NxHy[j]
        RxNxHy = rz * NxHx[j] - rx * NxHz[j]
        RxNxHz = rx * NxHy[j] - ry * NxHx[j]

        ie1x = (NxEx[j] - Z0 * RxNxHx) * G
        ie1y = (NxEy[j] - Z0 * RxNxHy) * G
        ie1z = (NxEz[j] - Z0 * RxNxHz) * G

        Eoutx += Q * (ry * ie1z - rz * ie1y)
        Eouty += Q * (rz * ie1x - rx * ie1z)
        Eoutz += Q * (rx * ie1y - ry * ie1x)

        # ii += iadd
        pgb.update(1)
    Eout[0, :] = Eoutx
    Eout[1, :] = Eouty
    Eout[2, :] = Eoutz

    Hout[0, :] = (ry * Eoutz - rz * Eouty) / Z0
    Hout[1, :] = (rz * Eoutx - rx * Eoutz) / Z0
    Hout[2, :] = (rx * Eouty - ry * Eoutx) / Z0

    return Eout, Hout

def stratton_chu(Ein, Hin, mesh: SurfaceMesh, theta: np.ndarray, phi: np.ndarray, k0: float):

    Ein = np.array(Ein)
    Hin = np.array(Hin)

    Emag = np.sqrt(np.abs(Ein[0,:])**2 + np.abs(Ein[1,:])**2 + np.abs(Ein[2,:])**2)
    Ntot = np.argwhere(Emag>0.000001*np.max(Emag)).shape[0]
    logger.debug(f'Percentage Included: {Ntot/Emag.shape[0]*100:.0f}%')
    areas = mesh.areas
    vis = mesh.edge_centers

    wns = np.zeros_like(vis).astype(np.float64)

    tri_normals = mesh.normals
    tri_ids = mesh.tri_to_edge

    for i in range(mesh.n_tris):
        n = tri_normals[:,i]
        i1, i2, i3 = tri_ids[:,i]
        wns[:,i1] += n*areas[i]/3
        wns[:,i2] += n*areas[i]/3
        wns[:,i3] += n*areas[i]/3
    
    Eout = None
    Hout = None
    tpout = np.array([theta, phi])
    with ProgressBar(total=Ntot, ncols=100, dynamic_ncols=False) as pgb:
        Eout, Hout = stratton_chu_ff(
            Ein.astype(np.complex128),
            Hin.astype(np.complex128),
            vis.astype(np.float64),
            wns.astype(np.float64),
            tpout.astype(np.float64),
            np.float64(k0),
            pgb,
        )
    return Eout.astype(np.complex128), Hout.astype(np.complex128)



# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

import numpy as np
from dataclasses import dataclass
from typing import Callable
@dataclass
class Material:
    """The Material class generalizes a material in the EMerge FEM environment.

    If a scalar value is provided for the relative permittivity or the relative permeability
    it will be used as multiplication entries for the material property diadic as identity matrix.

    Additionally, a function may be provided that computes a coordinate dependent material property
    for _fer. For example: Material(_fer = lambda x,y,z: ...). 
    The x,y and z coordinates are provided as a (N,) np.ndarray. The return array must be of shape (3,3,N)!
    
    """
    er: float = 1
    ur: float = 1
    tand: float = 0
    cond: float = 0
    _neff: float | None = None
    _fer: Callable | None= None
    _fur: Callable | None = None
    color: str = "#BEBEBE"
    _color_rgb: tuple[float,float,float] = (0.5, 0.5, 0.5)
    opacity: float = 1.0

    def __post_init__(self):
        hex_str = self.color.lstrip('#')
        self._color_rgb = tuple(int(hex_str[i:i+2], 16)/255.0 for i in (0, 2, 4))

    @property
    def sigma(self) -> float:
        return self.cond
        
    @property
    def color_rgb(self) -> tuple[float,float,float]:
        return self._color_rgb
    
    @property
    def ermat(self) -> np.ndarray:
        if isinstance(self.er, (float, complex, int, np.float64, np.complex128)):
            return self.er*(1-1j*self.tand)*np.eye(3)
        else:
            return self.er*(1-1j*self.tand)
    
    @property
    def urmat(self) -> np.ndarray:
        if isinstance(self.ur, (float, complex, int, np.float64, np.complex128)):
            return self.ur*np.eye(3)
        else:
            return self.ur
    
    @property
    def neff(self) -> complex:
        if self._neff is not None:
            return self._neff
        er = self.ermat[0,0]
        ur = self.urmat[0,0]

        return np.abs(np.sqrt(er*(1-1j*self.tand)*ur))
    
    @property
    def fer2d(self) -> Callable:
        if self._fer is None:
            return lambda x,y: self.er*(1-1j*self.tand)*np.ones_like(x)
        else:
            return self._fer
        
    @property
    def fur2d(self) -> Callable:
        if self._fur is None:

            return lambda x,y: self.ur*np.ones_like(x)
        else:
            return self._fur
    @property
    def fer3d(self) -> Callable:
        if self._fer is None:
            return lambda x,y,z: self.er*(1-1j*self.tand)*np.ones_like(x)
        else:
            return self._fer
    
    @property
    def fur3d(self) -> Callable:
        if self._fur is None:
            return lambda x,y,z: self.ur*np.ones_like(x)
        else:
            return self._fur
    @property
    def fer3d_mat(self) -> Callable:
        if self._fer is None:
            
            return lambda x,y,z: np.repeat(self.ermat[:, :, np.newaxis], x.shape[0], axis=2)
        else:
            return self._fer
    
    @property
    def fur3d_mat(self) -> Callable:
        if self._fur is None:
            return lambda x,y,z: np.repeat(self.urmat[:, :, np.newaxis], x.shape[0], axis=2)
        else:
            return self._fur
        
AIR = Material(color="#4496f3", opacity=0.05)
COPPER = Material(cond=5.8e7, color="#62290c")


# EMerge is an open source Python based FEM EM simulation module.
# Copyright (C) 2025  Robert Fennis.

# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <https://www.gnu.org/licenses/>.

from __future__ import annotations
import numpy as np
from ..mesh3d import Mesh3D
from .femdata import FEMBasis
from .ned2_interp import ned2_tet_interp, ned2_tet_interp_curl
from ..mth.optimized import local_mapping
from .index_interp import index_interp

############### Nedelec2 Class

class Nedelec2(FEMBasis):


    def __init__(self, mesh: Mesh3D):
        super().__init__(mesh)
        
        self.nedges: int = self.mesh.n_edges
        self.ntris: int = self.mesh.n_tris
        self.ntets: int = self.mesh.n_tets

        self.nfield: int = 2*self.nedges + 2*self.ntris
        
        ######## MESH Derived

        nedges = self.mesh.n_edges
        ntris = self.mesh.n_tris

        self.tet_to_field: np.ndarray = np.zeros((20, self.mesh.tets.shape[1]), dtype=int)
        self.tet_to_field[:6,:] = self.mesh.tet_to_edge
        self.tet_to_field[6:10,:] = self.mesh.tet_to_tri + nedges
        self.tet_to_field[10:16,:] = self.mesh.tet_to_edge + (ntris+nedges)
        self.tet_to_field[16:20,:] = self.mesh.tet_to_tri + (ntris+2*nedges)
    
        self.edge_to_field: np.ndarray = np.zeros((2,nedges), dtype=int)

        self.edge_to_field[0,:] = np.arange(nedges)
        self.edge_to_field[1,:] = np.arange(nedges) + ntris + nedges

        self.tri_to_field: np.ndarray = np.zeros((8,ntris), dtype=int)

        self.tri_to_field[:3,:] = self.mesh.tri_to_edge
        self.tri_to_field[3,:] = np.arange(ntris) + nedges
        self.tri_to_field[4:7,:] = self.mesh.tri_to_edge + nedges + ntris
        self.tri_to_field[7,:] = np.arange(ntris) + 2*nedges + ntris

        ##
        self._field: np.ndarray | None = None
        self.n_tet_dofs = 20
        self.n_tri_dofs = 8
        self._all_tet_ids = np.arange(self.ntets)

        self.empty_tri_rowcol()
    
    def interpolate(self, field: np.ndarray, xs: np.ndarray, ys: np.ndarray, zs:np.ndarray, tetids: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        ''' 
        Interpolate the provided field data array at the given xs, ys and zs coordinates
        '''
        if tetids is None:
            tetids = self._all_tet_ids
        return ned2_tet_interp(np.array([xs, ys, zs]), field, self.mesh.tets, self.mesh.tris, self.mesh.edges, self.mesh.nodes, self.tet_to_field, tetids)
    
    def interpolate_curl(self, field: np.ndarray, xs: np.ndarray, ys: np.ndarray, zs:np.ndarray, c: np.ndarray, tetids: np.ndarray | None = None) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Interpolates the curl of the field at the given points.
        """
        if tetids is None:
            tetids = self._all_tet_ids
        return ned2_tet_interp_curl(np.array([xs, ys, zs]), field, self.mesh.tets, self.mesh.tris, self.mesh.edges, self.mesh.nodes, self.tet_to_field, c, tetids)
    
    def interpolate_index(self, xs: np.ndarray,
                        ys: np.ndarray,
                        zs: np.ndarray,
                        tetids: np.ndarray | None = None) -> np.ndarray:
        if tetids is None:
            tetids = self._all_tet_ids

        return index_interp(np.array([xs, ys, zs]), self.mesh.tets, self.mesh.nodes, tetids)
    ###### INDEX MAPPINGS

    def local_tet_to_triid(self, itet: int) -> np.ndarray:
        tri_ids = self.tet_to_field[6:10, itet] - self.n_edges
        global_tri_map = self.mesh.tris[:, tri_ids]
        return local_mapping(self.mesh.tets[:, itet], global_tri_map)

    def local_tet_to_edgeid(self, itet: int) -> np.ndarray:
        global_edge_map = self.mesh.edges[:, self.tet_to_field[:6,itet]]
        return local_mapping(self.mesh.tets[:, itet], global_edge_map)

    def local_tri_to_edgeid(self, itri: int) -> np.ndarray:
        global_edge_map = self.mesh.edges[:, self.tri_to_field[:3,itri]]
        return local_mapping(self.mesh.tris[:, itri], global_edge_map)
    
    def map_edge_to_field(self, edge_ids: np.ndarray) -> np.ndarray:
        """
        Returns the field ids for the edges.
        """
        # Concatinate the edges with the edges + ntris + nedges
        edge_ids = np.array(edge_ids)
        return np.concatenate((edge_ids, edge_ids + self.ntris + self.nedges))
    
    ########
    # @staticmethod
    # def tet_stiff_mass_submatrix(tet_vertices: np.ndarray, 
    #                              edge_lengths: np.ndarray, 
    #                              local_edge_map: np.ndarray, 
    #                              local_tri_map: np.ndarray, 
    #                              C_stiffness: float, 
    #                              C_mass: float) -> tuple[np.ndarray, np.ndarray]:
    #     return ned2_tet_stiff_mass(tet_vertices, edge_lengths, local_edge_map, local_tri_map, C_stiffness, C_mass)
    
    # @staticmethod
    # def tri_stiff_vec_matrix(lcs_vertices: np.ndarray, 
    #                          gamma: complex, 
    #                          lcs_Uinc: np.ndarray, 
    #                          DPTs: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
    #     return ned2_tri_stiff_force(lcs_vertices, gamma, lcs_Uinc, DPTs)

    # @staticmethod
    # def tri_surf_integral(lcs_vertices: np.ndarray, 
    #                       edge_lengths: np.ndarray, 
    #                       lcs_Uinc: np.ndarray, 
    #                       DPTs: np.ndarray) -> complex:
    #     return ned2_tri_surface_integral(lcs_vertices, edge_lengths, lcs_Uinc, DPTs)


import emerge as em
import numpy as np
from emerge.plot import plot_sp, smith, plot_ff_polar, plot_ff

""" PATCH ANTENNA DEMO

This design is modeled after this Comsol Demo: https://www.comsol.com/model/microstrip-patch-antenna-11742

In this demo we build and simulate a rectangular patch antenna on a dielectric
substrate with airbox and lumped port excitation, then visualize S-parameters
and far-field radiation patterns. """

# --- Unit and simulation parameters --------------------------------------
mm = 0.001              # meters per millimeter
margin = 5 * mm         # extra air margin around substrate
Nmodes = 1              # number of port modes to excite

# --- Antenna geometry dimensions ----------------------------------------
Wpatch = 53 * mm        # patch width (meters)
Lpatch = 52 * mm        # patch length
wline = 3.2 * mm        # feed line width
wstub = 7 * mm          # stub width for inset feed
lstub = 15.5 * mm       # stub (feed) length
wsub = 100 * mm         # substrate width
hsub = 100 * mm         # substrate length
th = 1.524 * mm         # substrate thickness

# --- Airbox and dielectric constants ------------------------------------
Hair = 40 * mm          # airbox height above substrate
er = 3.38               # substrate dielectric constant (relative)

# Refined frequency range for antenna resonance around 1.54–1.6 GHz
f1 = 1.54e9             # start frequency
f2 = 1.6e9              # stop frequency

# --- Create simulation object -------------------------------------------
# Using PVDisplay backend for 3D visualization
model = em.Simulation3D('MyPatchAntenna', loglevel='DEBUG')

# --- Define geometry primitives -----------------------------------------
# Substrate block centered at origin in XY, thickness in Z (negative down)
dielectric = em.geo.Box(wsub, hsub, th,
                        position=(-wsub/2, -hsub/2, -th))
# Air box above substrate (Z positive)
air = em.geo.Box(wsub, hsub, Hair,
                  position=(-wsub/2, -hsub/2, 0))

# Metal patch rectangle on top of substrate
rpatch = em.geo.XYPlate(Wpatch, Lpatch,
                        position=(-Wpatch/2, -Lpatch/2, 0))

# Define cutouts for inset feed: two rectangular plates to subtract
cutout1 = em.geo.XYPlate(wstub, lstub,
                         position=(-wline/2 - wstub, -Lpatch/2, 0))
cutout2 = em.geo.XYPlate(wstub, lstub,
                         position=( wline/2, -Lpatch/2, 0))
# Feed line plate to add back between cutouts
line = em.geo.XYPlate(wline, lstub,
                       position=(-wline/2, -Lpatch/2, 0))
# Plate defining lumped port geometry (origin + width/height vectors)
port = em.geo.Plate(
    np.array([-wline/2, -Lpatch/2, -th]),  # lower port corner
    np.array([wline, 0, 0]),                # width vector along X
    np.array([0, 0, th])                    # height vector along Z
)

# Build final patch shape: subtract cutouts, add feed line
rpatch = em.geo.remove(rpatch, cutout1)
rpatch = em.geo.remove(rpatch, cutout2)
rpatch = em.geo.add(rpatch, line)

# Assign copper material for visualization only
rpatch.material = em.lib.MET_COPPER

# --- Assign materials and simulation settings ---------------------------
# Dielectric material with some transparency for display
dielectric.material = em.Material(er, tand=0.01,
                                  color="#207020", opacity=0.6)
# Mesh resolution: fraction of wavelength
model.mw.resolution = 0.2
# Frequency sweep across the resonance
model.mw.set_frequency_range(f1, f2, 21)

# --- Combine geometry into simulation -----------------------------------
model.commit_geometry()

# --- Mesh refinement settings --------------------------------------------
# Finer boundary mesh on patch edges for accuracy
model.mesher.set_boundary_size(rpatch, 5 * mm, 1.1)
# Refined mesh on port face for excitation accuracy
model.mesher.set_face_size(port, 0.5 * mm)

# --- Generate mesh and preview ------------------------------------------
model.generate_mesh()                      # build the finite-element mesh
model.view(selections=[port])              # show the mesh around the port

# --- Boundary conditions ------------------------------------------------
# Define lumped port with specified orientation and impedance
port = model.mw.bc.LumpedPort(
    port, 1,
    width=wline, height=th,
    direction=em.ZAX,
    active=True, Z0=50
)
# Apply absorbing boundary on underside of airbox to simulate open space
boundary_selection = air.outside('bottom')

model.view(selections=[boundary_selection,])
abc = model.mw.bc.AbsorbingBoundary(boundary_selection)
# Perfect conductor on the metallic patch
pec = model.mw.bc.PEC(rpatch)

# --- Run frequency-domain solver ----------------------------------------
data = model.mw.frequency_domain()

# --- Post-process S-parameters ------------------------------------------
freqs = data.scalar.grid.freq
S11 = data.scalar.grid.S(1, 1)            # reflection coefficient
plot_sp(freqs / 1e9, S11)                 # plot return loss in dB
smith(freqs, S11)                         # Smith chart of S11

# --- Far-field radiation pattern ----------------------------------------
# Extract 2D cut at phi=0 plane and plot E-field magnitude
theta, Exax, Hxax = data.field.find(freq=1.56e9)\
    .farfield_2d((0, 0, 1), (1, 0, 0), boundary_selection)
theta, Eyax, Hyax = data.field.find(freq=1.56e9)\
    .farfield_2d((0, 0, 1), (0, 1, 0), boundary_selection)

plot_ff(theta, [em.norm(Exax), em.norm(Eyax)])                # linear plot vs theta
plot_ff_polar(theta, [em.norm(Exax), em.norm(Eyax)])          # polar plot of radiation

# --- 3D radiation visualization -----------------------------------------
# Add geometry to 3D display
model.display.add_object(rpatch)
model.display.add_object(dielectric)
# Compute full 3D far-field and display surface colored by |E|
ff3d = data.field[1].farfield_3d(abc)
surf = ff3d.surfplot('normE', rmax=60 * mm,
                      offset=(0, 0, 20 * mm))
model.display.add_surf(*surf, cmap='viridis', symmetrize=False)
model.display.show()